import json
import logging
from .records_format_file import RecordsFormatFile
from .records_schema_sql_file import RecordsSchemaSqlFile
from .records_schema_json_file import RecordsSchemaJsonFile
from .schema import RecordsSchema
from urllib.parse import urlparse
from ..url.base import BaseDirectoryUrl, BaseFileUrl
from .records_format import BaseRecordsFormat, DelimitedRecordsFormat
from typing import Mapping, IO, List, Optional
from .records_types import UrlDetails, RecordsManifestWithLength, LegacyRecordsManifest


logger = logging.getLogger(__name__)


class RecordsDirectory:
    def __init__(self,
                 records_loc: BaseDirectoryUrl) -> None:
        self.loc = records_loc
        self.scheme = records_loc.scheme
        self.schema_sql_file = RecordsSchemaSqlFile(self.loc)
        self.schema_json_file = RecordsSchemaJsonFile(self.loc)
        self.format_file = RecordsFormatFile(self.loc)

    def save_schema(self, records_schema: RecordsSchema) -> None:
        json = records_schema.to_json()
        self.schema_json_file.save_schema_json(json)

    def load_schema_sql_from_sql_file(self) -> Optional[str]:
        return self.schema_sql_file.load_schema_sql()

    def load_schema_json(self) -> Optional[str]:
        return self.schema_json_file.load_schema_json()

    def load_schema_json_obj(self) -> Optional[RecordsSchema]:
        json = self.load_schema_json()
        if json is None:
            return None
        else:
            return RecordsSchema.from_json(json)

    def save_format(self, records_format: BaseRecordsFormat) -> None:
        self.format_file.save_format(records_format)

    def load_format(self, fail_if_dont_understand: bool) -> BaseRecordsFormat:
        return self.format_file.load_format(fail_if_dont_understand)

    def save_fileobjs(self,
                      fileobjs_by_target_names: Mapping[str, IO[bytes]],
                      records_schema: Optional[RecordsSchema]=None,
                      records_format: Optional[BaseRecordsFormat]=None) \
            -> UrlDetails:
        """Write out a full records directory from file objects."""
        url_details: UrlDetails = self.save_data_from_fileobjs(fileobjs_by_target_names)
        self.save_preliminary_manifest(url_details)
        if records_schema:
            self.save_schema(records_schema)
        if records_format:
            self.save_format(records_format)
        self.finalize_manifest()
        # Help ensure that future use of this directory will succeed
        self.loc.file_in_this_directory('_manifest').wait_to_exist()
        return url_details

    def save_data_from_fileobjs(self,
                                fileobjs_by_target_names: Mapping[str,
                                                                  IO[bytes]]) -> UrlDetails:
        """
        Write out just the datafiles into the records directory.

        Prefer save_fileobjs when writing a complete records directory.
        """
        url_details: UrlDetails = {}
        for target_name, fileobj in fileobjs_by_target_names.items():
            target_loc = self.loc.file_in_this_directory(target_name)
            logger.info(f"Uploading {target_loc.url}")
            length: int = target_loc.upload_fileobj(fileobj)
            url_details[target_loc.url] = {
                'content_length': length,
            }
        return url_details

    def _build_manifest(self,
                        url_details: UrlDetails) -> RecordsManifestWithLength:
        return {
            "entries": [
                {
                    "url": url,
                    "mandatory": True,
                    "meta": details
                }
                for url, details in url_details.items()
            ]
        }

    def save_preliminary_manifest(self, url_details: Optional[UrlDetails]=None) -> None:
        """Save manifest file to records directory.

         :param urls_for_manifest: list of URLs that should be
          listed in manifest. Defaults to None, in which case it lists
          files in directory (which is generally not advised because
          it could capture unintentional files).
        """

        manifest_loc = self.loc.file_in_this_directory('manifest')
        if url_details is None:
            logger.warning("Building manifest by listing directory contents")
            url_details = {
                loc.url: {
                    'content_length': loc.size()
                }
                for loc in self.loc.files_in_directory()
                if not loc.filename().startswith('_') and loc.filename() != 'manifest'
            }
        logger.info(f"Storing manifest into {manifest_loc.url}")
        manifest_loc.store_string(json.dumps(self._build_manifest(url_details)))

    def _filename_of_url(self, url: str) -> str:
        path = urlparse(url).path
        return path.split('/')[-1]

    def copy_to(self, new_loc: BaseDirectoryUrl) -> 'RecordsDirectory':
        logger.info(f"Copying files from {self.loc} to {new_loc}...")
        new_loc = self.loc.copy_to(new_loc)
        # rebuild manifest to point to new URLs.
        new_directory = RecordsDirectory(records_loc=new_loc)
        # regenerate manifest with new URLs
        old_urls: List[str] = self.manifest_entry_urls()
        new_urls: UrlDetails = {
            old_loc.url: {
                'content_length': old_loc.size()
            }
            for old_url
            in old_urls
            for old_loc
            in [new_loc.file_in_this_directory(self._filename_of_url(old_url))]
        }
        new_directory.save_preliminary_manifest(new_urls)
        old_finalized_manifest = self.loc.file_in_this_directory('_manifest')
        if old_finalized_manifest.exists():
            new_directory.finalize_manifest()
        return new_directory

    def copy_from(self, old_loc: BaseDirectoryUrl) -> 'RecordsDirectory':
        old_directory = RecordsDirectory(records_loc=old_loc)
        return old_directory.copy_to(self.loc)

    def finalize_manifest(self) -> None:
        old_loc = self.loc.file_in_this_directory('manifest')
        new_loc = self.loc.file_in_this_directory('_manifest')
        if new_loc.exists():
            logger.info(f"Deleting {new_loc} and finalizing new manifest in its place")
            new_loc.delete()
        old_loc.rename_to(new_loc)
        new_loc.wait_to_exist()

    def get_manifest(self) -> LegacyRecordsManifest:
        """Note that this returns a 'LegacyRecordsManifest' type - the output
        is not guaranteed to contain file length entries, as entries
        may have been written out by older versions of Records Mover
        prior to the file lengths being added to the manifest.
        """
        manifest_loc = self.loc.file_in_this_directory('_manifest')
        if not manifest_loc.exists():
            manifest_loc = self.loc.file_in_this_directory('manifest')
        return manifest_loc.json_contents()  # type: ignore

    def manifest_entry_urls(self) -> List[str]:
        manifest = self.get_manifest()
        if manifest is None:
            raise TypeError('Manifest empty')
        else:
            return [entry['url'] for entry in manifest['entries']]

    def save_to_url(self, output_loc: BaseFileUrl) -> None:
        records_format = self.load_format(fail_if_dont_understand=True)
        manifest_entry_urls = self.manifest_entry_urls()
        if len(manifest_entry_urls) == 1:
            input_filename = self._filename_of_url(manifest_entry_urls[0])
            input_loc = self.loc.file_in_this_directory(input_filename)
            input_loc.copy_to(output_loc)
        else:
            if isinstance(records_format, DelimitedRecordsFormat):
                if records_format.hints['header-row']:
                    raise NotImplementedError("Please teach me how to concatenate "
                                              "delimited files with header row")
                else:
                    input_filenames = [self._filename_of_url(input_url)
                                       for input_url in manifest_entry_urls]
                    input_locs = [self.loc.file_in_this_directory(input_filename)
                                  for input_filename in input_filenames]
                    logger.info(f"Saving files from {self.loc.url} to {output_loc.url}")
                    output_loc.concatenate_from(input_locs)
            else:
                raise NotImplementedError("Please teach me how to concatenate this format of file")

    def await_completion(self,
                         manifest_filename: str = "_manifest",
                         log_level: int = logging.INFO,
                         ms_between_polls: int = 50) -> None:
        manifest_loc = self.loc.file_in_this_directory(manifest_filename)
        manifest_loc.wait_to_exist(log_level=log_level, ms_between_polls=ms_between_polls)
        manifest_locs = [self.loc.file_in_this_directory(self._filename_of_url(url))
                         for url in self.manifest_entry_urls()]
        for loc in manifest_locs:
            loc.wait_to_exist(log_level=log_level, ms_between_polls=ms_between_polls)

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.loc.url})"



import logging
from records_mover.mover_types import _ensure_all_cases_covered
from .processing_instructions import ProcessingInstructions
from . import PartialRecordsHints, UntypedRecordsHints
from .base_records_format import BaseRecordsFormat
from typing import Mapping, Optional, TYPE_CHECKING
from .delimited import ValidatedRecordsHints
if TYPE_CHECKING:
    from . import RecordsFormatType  # noqa

logger = logging.getLogger(__name__)


class AvroRecordsFormat(BaseRecordsFormat):
    "Describes records files in `Avro <https://avro.apache.org/>`_ format"

    def __init__(self) -> None:
        "Create a new instance of AvroRecordsFormat"
        self.format_type = 'avro'

    def __str__(self) -> str:
        return "AvroRecordsFormat"

    def __repr__(self) -> str:
        return str(self)

    def generate_filename(self, basename: str) -> str:
        return f"{basename}.avro"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, AvroRecordsFormat):
            return True
        return False


class ParquetRecordsFormat(BaseRecordsFormat):
    "Describes records files in `Parquet <https://parquet.apache.org/>`_ format"

    def __init__(self) -> None:
        "Create a new instance of ParquetRecordsFormat"
        self.format_type = 'parquet'

    def __str__(self) -> str:
        return "ParquetRecordsFormat"

    def __repr__(self) -> str:
        return str(self)

    def generate_filename(self, basename: str) -> str:
        return f"{basename}.parquet"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ParquetRecordsFormat):
            return True
        return False


class DelimitedRecordsFormat(BaseRecordsFormat):
    "Describes records data files in delimited (CSV) format."

    variant: str
    # The strong type for 'hints' in the constructor is for IDE
    # support--we don't actually trust input to this class to be well-typed,
    # as it may come from public interface consumers.
    hints: UntypedRecordsHints
    """Stores the full set of hints describing the format, combining both
    the default hints for the variant and any hint overrides provided
    in the constructor"""

    def __init__(self,
                 variant: str='bluelabs',
                 hints: PartialRecordsHints={},
                 processing_instructions: ProcessingInstructions=ProcessingInstructions()) -> None:
        """See the `records format documentation
        <https://github.com/bluelabsio/records-mover/blob/master/docs/RECORDS_SPEC.md#hints>`_
        for full details on parameters.

        :param variant: For a given type (especially delimited), describe the subtype of the
           format.  For 'delimited', valid values include 'dumb', 'csv', 'bluelabs', 'bigquery'
           and 'vertica'.
        :param hints: Dictionary of names of delimited hints mapping to their values.  See the
           `records format specification
           <https://github.com/bluelabsio/records-mover/blob/master/docs/RECORDS_SPEC.md>`_
           for hints and valid values.
        :param processing_instructions: Directives on how to handle different situations when
           processing files.
        :type processing_instructions: records_mover.records.ProcessingInstructions
        """
        self.format_type = 'delimited'
        self.variant = variant
        self.custom_hints = hints
        self.add_hints_from_variant(provided_hints=hints,
                                    processing_instructions=processing_instructions)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DelimitedRecordsFormat):
            return (self.format_type == other.format_type and
                    self.variant == other.variant and
                    self.hints == other.hints)
        return False

    def alter_hints(self, new_hints: UntypedRecordsHints) ->\
            'DelimitedRecordsFormat':
        input_hints = dict(self.custom_hints)  # make copy
        input_hints.update(new_hints)
        return DelimitedRecordsFormat(variant=self.variant,
                                      hints=input_hints)  # type: ignore

    def alter_variant(self, variant: str) -> 'DelimitedRecordsFormat':
        return DelimitedRecordsFormat(variant=variant,
                                      hints=self.custom_hints)

    def base_hints_from_variant(self,
                                fail_if_dont_understand: bool = True) -> PartialRecordsHints:
        hint_defaults: PartialRecordsHints = {
            'header-row': False,
            'field-delimiter': ',',
            'record-terminator': "\n",
            'compression': 'GZIP',
            'quoting': None,
            'quotechar': '"',
            'doublequote': False,
            'escape': None,
            'encoding': 'UTF8',
            'dateformat': 'YYYY-MM-DD',
            'timeonlyformat': 'HH24:MI:SS',
            'datetimeformat': 'YYYY-MM-DD HH:MI:SS',
            'datetimeformattz': 'YYYY-MM-DD HH:MI:SSOF',
        }
        combined_hints: PartialRecordsHints = dict(hint_defaults)  # type: ignore
        format_driven_hints: PartialRecordsHints = {}  # noqa
        if self.variant == 'dumb':
            format_driven_hints['field-delimiter'] = ','
            format_driven_hints['record-terminator'] = "\n"
            format_driven_hints['quoting'] = None
            format_driven_hints['doublequote'] = False
        elif self.variant == 'csv':
            format_driven_hints['field-delimiter'] = ','
            format_driven_hints['quoting'] = 'minimal'
            format_driven_hints['doublequote'] = True
            format_driven_hints['quotechar'] = '"'
            format_driven_hints['dateformat'] = 'MM/DD/YY'
            format_driven_hints['datetimeformat'] = 'MM/DD/YY HH24:MI'
            format_driven_hints['datetimeformattz'] = 'MM/DD/YY HH24:MI'
            format_driven_hints['timeonlyformat'] = 'HH24:MI:SS'
            format_driven_hints['header-row'] = True
        elif self.variant == 'bigquery':
            format_driven_hints['field-delimiter'] = ','
            format_driven_hints['quoting'] = 'minimal'
            format_driven_hints['doublequote'] = True
            format_driven_hints['quotechar'] = '"'
            format_driven_hints['header-row'] = True
            format_driven_hints['datetimeformat'] = 'YYYY-MM-DD HH:MI:SS'
            format_driven_hints['datetimeformattz'] = 'YYYY-MM-DD HH:MI:SS'
        elif self.variant == 'bluelabs':
            format_driven_hints['field-delimiter'] = ','
            format_driven_hints['escape'] = '\\'
            format_driven_hints['quoting'] = None
            format_driven_hints['doublequote'] = False
            format_driven_hints['dateformat'] = 'YYYY-MM-DD'
            format_driven_hints['datetimeformat'] = 'YYYY-MM-DD HH24:MI:SS'
        elif self.variant == 'vertica':
            format_driven_hints['field-delimiter'] = '\001'
            format_driven_hints['record-terminator'] = '\002'
            format_driven_hints['quoting'] = None
            format_driven_hints['doublequote'] = False
            format_driven_hints['escape'] = None
            format_driven_hints['compression'] = None
            format_driven_hints['timeonlyformat'] = 'HH24:MI:SS'
        else:
            if fail_if_dont_understand:
                raise NotImplementedError("Implement delimited records format variant "
                                          f"{self.variant} or try "
                                          "again with fail_if_dont_understand=False")
            else:
                logger.warning("Ignoring delimited records format "
                               f"variant {self.variant}")
        combined_hints.update(format_driven_hints)
        return combined_hints

    def add_hints_from_variant(self,
                               provided_hints: PartialRecordsHints,
                               processing_instructions: ProcessingInstructions) -> None:
        combined_hints =\
            self.base_hints_from_variant(processing_instructions.fail_if_dont_understand)
        combined_hints.update(provided_hints)  # user's hints override those from the format
        self.hints = combined_hints

    def config(self) -> Optional[Mapping[str, object]]:
        contents_data = None
        if self.variant is not None or len(self.hints) != 0:
            contents_data = {
                'type': self.format_type,
                'variant': self.variant,
                'hints': self.hints
            }
        return contents_data

    def __str__(self) -> str:
        hints_from_variant = self.base_hints_from_variant()
        hint_overrides = {
            hint: v for hint, v in self.hints.items()
            if hint not in hints_from_variant or v != hints_from_variant[hint]  # type: ignore
        }
        if hint_overrides != {}:
            return f"DelimitedRecordsFormat({self.variant} - {hint_overrides})"
        else:
            return f"DelimitedRecordsFormat({self.variant})"

    def generate_filename(self, basename: str) -> str:
        compression_type = self.hints['compression']
        if compression_type is None:
            return f"{basename}.csv"
        elif compression_type == 'GZIP':
            return f"{basename}.csv.gz"
        elif compression_type == 'BZIP':
            return f"{basename}.csv.bz2"
        else:
            raise NotImplementedError("Teach me how to handle compression "
                                      f"type {compression_type}")

    def validate(self,
                 fail_if_cant_handle_hint: bool) -> ValidatedRecordsHints:
        return ValidatedRecordsHints.validate(self.hints,
                                              fail_if_cant_handle_hint=fail_if_cant_handle_hint)

    def __repr__(self) -> str:
        return str(self)


def RecordsFormat(format_type: 'RecordsFormatType' = 'delimited',
                  variant: str='bluelabs',
                  hints: PartialRecordsHints={},
                  processing_instructions:
                  ProcessingInstructions=ProcessingInstructions()) ->\
            'BaseRecordsFormat':
    if format_type == 'delimited':
        return DelimitedRecordsFormat(variant=variant,
                                      hints=hints,
                                      processing_instructions=processing_instructions)
    elif format_type == 'parquet':
        return ParquetRecordsFormat()
    elif format_type == 'avro':
        return AvroRecordsFormat()
    else:
        _ensure_all_cases_covered(format_type)
        raise NotImplementedError(f'Teach me to handle format type {format_type}')



import pathlib
from ..records_format import BaseRecordsFormat
from ..schema import RecordsSchema
from ..records_directory import RecordsDirectory
from ..processing_instructions import ProcessingInstructions
from ...url.resolver import UrlResolver
from .fileobjs import FileobjsSource  # noqa
from .uninferred_fileobjs import UninferredFileobjsRecordsSource
from .data_url import DataUrlRecordsSource
from .directory import RecordsDirectoryRecordsSource
from .. import PartialRecordsHints
from .base import (SupportsRecordsDirectory, SupportsMoveToRecordsDirectory,  # noqa
                   SupportsToFileobjsSource, RecordsSource)
from typing import Mapping, IO, Callable, Optional, Union, Iterable, TYPE_CHECKING
if TYPE_CHECKING:
    # see the 'gsheets' extras_require option in setup.py - needed for this!
    import google.auth.credentials  # noqa
    from sqlalchemy.engine import Engine  # noqa
    from ...db import DBDriver  # noqa
    from .google_sheets import GoogleSheetsRecordsSource  # noqa ditto
    # with pandas, which an optional addition for clients of this
    # library
    from pandas import DataFrame  # noqa
    from .dataframes import DataframesRecordsSource  # noqa
    from .table import TableRecordsSource  # noqa


class RecordsSources(object):
    """These methods produce objects representing the source of a records
    move.  The objects can be used as the 'source' argument to
    :meth:`records_mover.records.move`

    This object should be pulled from the 'sources' property of the
    'records' property on a :class:`records_mover.Session` object
    instead of being constructed directly.

    Example use:

    .. code-block:: python

       records = session.records
       db_engine = session.get_default_db_engine()
       url = 's3://some-bucket/some-directory/'
       source = records.sources.directory_from_url(url=url)
       target = records.targets.table(schema_name='myschema',
                                      table_name='mytable',
                                      db_engine=db_engine)
       results = records.move(source, target)

    """
    def __init__(self,
                 db_driver: Callable[['Engine'], 'DBDriver'],
                 url_resolver: UrlResolver) -> None:
        self.db_driver = db_driver
        self.url_resolver = url_resolver

    def dataframe(self,
                  df: 'DataFrame',
                  processing_instructions: ProcessingInstructions=
                  ProcessingInstructions(),
                  records_schema: Optional[RecordsSchema]=None,
                  include_index: bool=False) -> 'DataframesRecordsSource':
        """
        Represents a single dataframe source.

        :param df: Pandas DataFrame to move data from.
        :param processing_instructions: Instructions used during creation of the schema SQL as a
           :class:`records_mover.records.ProcessingInstructions` object.
        :param include_index: If True, the Pandas DataFrame index column will be included in the
           move as a column; if False, it will be disregarded.
        :param records_schema: Experimental interface; do not use.
        """

        from .dataframes import DataframesRecordsSource  # noqa
        return DataframesRecordsSource(dfs=[df],
                                       records_schema=records_schema,
                                       processing_instructions=processing_instructions,
                                       include_index=include_index)

    def dataframes(self,
                   dfs: Iterable['DataFrame'],
                   processing_instructions: ProcessingInstructions=
                   ProcessingInstructions(),
                   records_schema: Optional[RecordsSchema]=None,
                   include_index: bool=False) -> 'DataframesRecordsSource':
        """Represents multiple dataframes as a source.  Note that this accepts
        an iterable, meaning that the dataframes in question can be
        generated dynamically in chunks.

        :param dfs: Iterable of Pandas DataFrames to move data from -- all data from these
           DataFrames will be added to the same table.
        :param processing_instructions: Instructions used during creation of the schema SQL as a
           :class:`records_mover.records.ProcessingInstructions` object.
        :param include_index: If True, the Pandas DataFrame index column will be included in the
           move as a column; if False, it will be disregarded.
        :param records_schema: Experimental interface; do not use.
        """
        from .dataframes import DataframesRecordsSource  # noqa
        return DataframesRecordsSource(dfs=dfs,
                                       records_schema=records_schema,
                                       processing_instructions=processing_instructions,
                                       include_index=include_index)

    def fileobjs(self,
                 target_names_to_input_fileobjs: Mapping[str, IO[bytes]],
                 records_format: Optional[BaseRecordsFormat]=None,
                 initial_hints: Optional[PartialRecordsHints]=None,
                 records_schema: Optional[RecordsSchema]=None)\
            -> Union[UninferredFileobjsRecordsSource, FileobjsSource]:
        """Represents one or more streams of data files as a source.

        :param target_names_to_input_fileobjs: Filenames mapping to streams of data file.
        :param records_format: Description of the format of the data files.
        :param initial_hints: If records_format is not provided, the format of the file will be
           determined automatically.  If that effort fails, you can help it out by providing hints
           in this dictionary as needed.  See the `records format specification
           <https://github.com/bluelabsio/records-mover/blob/master/docs/RECORDS_SPEC.md>`_ for
           hints and valid values.
        :param records_schema: Experimental interface; do not use.
        """
        if records_schema is None or records_format is None:
            return UninferredFileobjsRecordsSource(
                target_names_to_input_fileobjs=target_names_to_input_fileobjs,
                records_format=records_format,
                records_schema=records_schema,
                initial_hints=initial_hints)
        else:
            return FileobjsSource(
                target_names_to_input_fileobjs=target_names_to_input_fileobjs,
                records_format=records_format,
                records_schema=records_schema)

    def data_url(self,
                 input_url: str,
                 records_format: Optional[BaseRecordsFormat]=None,
                 initial_hints: Optional[PartialRecordsHints]=None,
                 records_schema: Optional[RecordsSchema]=None)\
            -> DataUrlRecordsSource:
        """Represents a URL pointer to a data file as a source.

        :param input_url: Location of the data file.  Must be a URL format understood by the
           records_mover.url library.
        :param records_format: Description of the format of the data files.
        :param initial_hints: If records_format is not provided, the format of the file will be
           determined automatically.  If that effort fails, you can help it out by providing hints
           in this dictionary as needed.  See the `records format specification
           <https://github.com/bluelabsio/records-mover/blob/master/docs/RECORDS_SPEC.md>`_ for
           hints and valid values.
        :param records_schema: Experimental interface; do not use.
        """
        return DataUrlRecordsSource(input_url=input_url,
                                    url_resolver=self.url_resolver,
                                    records_format=records_format,
                                    records_schema=records_schema,
                                    initial_hints=initial_hints)

    def table(self,
              db_engine: 'Engine',
              schema_name: str,
              table_name: str) -> 'TableRecordsSource':
        """Represents a SQLALchemy-accessible database table as as a source.

        :param db_engine: SQLAlchemy database engine to pull data from.
        :param schema_name: Schema name of a table to get data from.
        :param table_name: Table name of a table to get data from.
        """
        from .table import TableRecordsSource  # noqa
        return TableRecordsSource(schema_name=schema_name,
                                  table_name=table_name,
                                  url_resolver=self.url_resolver,
                                  driver=self.db_driver(db_engine))

    def directory_from_url(self,
                           url: str,
                           hints: PartialRecordsHints={},
                           fail_if_dont_understand: bool=True)\
            -> RecordsDirectoryRecordsSource:
        """Represents a Records Directory pointed to by a URL as a source.

        :param url: Location of the records directory.  Must be a URL format understood by the
           records_mover.url library, and must be a directory URL that ends with a '/'.
        :param hints: Any additional hints that should override the description of the data files
           already in the records directory.
        :param fail_if_dont_understand: If True, and a part of the RecordsFormat is not understood
           while processing, then immediately fail and raise an exception.  Otherwise, ignore the
           misunderstood instruction (e.g., ignore the hint, assume default variant, etc etc)
        """
        directory_loc = self.url_resolver.directory_url(url)
        directory = RecordsDirectory(records_loc=directory_loc)
        fail = fail_if_dont_understand
        return RecordsDirectoryRecordsSource(directory=directory,
                                             fail_if_dont_understand=fail,
                                             override_hints=hints,
                                             url_resolver=self.url_resolver)

    def local_file(self,
                   filename: str,
                   records_format: Optional[BaseRecordsFormat]=None,
                   initial_hints: Optional[PartialRecordsHints]=None,
                   records_schema: Optional[RecordsSchema]=None)\
            -> DataUrlRecordsSource:
        """Represents a data file on the local filesystem as a source.

        :param filename: File path (relative or absolute) of the data file to load.
        :param records_format: Description of the format of the data files.
        :param initial_hints: If records_format is not provided, the format of the file will be
           determined automatically.  If that effort fails, you can help it out by providing hints
           in this dictionary as needed.  See the `records format specification
           <https://github.com/bluelabsio/records-mover/blob/master/docs/RECORDS_SPEC.md>`_ for
           hints and valid values.
        """
        url = pathlib.Path(filename).resolve().as_uri()
        return self.data_url(input_url=url,
                             records_format=records_format,
                             records_schema=records_schema,
                             initial_hints=initial_hints)

    def google_sheet(self,
                     spreadsheet_id: str,
                     sheet_name_or_range: str,
                     google_cloud_creds:
                     'google.auth.credentials.Credentials',
                     out_of_band_column_headers: Optional[Iterable[str]]=None,
                     header_translator: Optional[Callable[[str], str]]=None) ->\
            'GoogleSheetsRecordsSource':
        """Represents a sheet or range in a Google Sheets spreadsheet as a
        source, via the Google Sheets API.

        :param spreadsheet_id: This is the xyz in
           https://docs.google.com/spreadsheets/d/xyz/edit?ts=5be5b383#gid=abc
        :param sheet_name_or_range: This is the label of the particular tab within the Google
           Sheets spreadsheet where the data should go, or a valid Google Sheets-style range formula
        :param google_cloud_creds: This is an object representing Google Cloud Platform access
           credentials.
        :param out_of_band_column_headers: If provided, we'll use these column names instead of the
           first row of the spreadsheet.  If set, the first row will be treated as data.
        :param header_translator: If provided, header names pulled from the sheet will be
           translated through this function.  Not used if out_of_band_column_headers is set.
        """
        from .google_sheets import GoogleSheetsRecordsSource  # noqa
        return GoogleSheetsRecordsSource(spreadsheet_id=spreadsheet_id,
                                         sheet_name_or_range=sheet_name_or_range,
                                         google_cloud_creds=google_cloud_creds,
                                         out_of_band_column_headers=out_of_band_column_headers,
                                         header_translator=header_translator)



import sqlalchemy
import sqlalchemy.dialects.mysql
import logging
from ...utils.limits import (IntegerType,
                             FLOAT32_SIGNIFICAND_BITS,
                             FLOAT64_SIGNIFICAND_BITS,
                             num_digits)
from ..driver import DBDriver
from .loader import MySQLLoader
from typing import Optional, Tuple, Union
from ..loader import LoaderFromFileobj, LoaderFromRecordsDirectory
from ...url.resolver import UrlResolver


logger = logging.getLogger(__name__)


class MySQLDBDriver(DBDriver):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection],
                 url_resolver: UrlResolver,
                 **kwargs) -> None:
        super().__init__(db)
        self._mysql_loader = MySQLLoader(db=db,
                                         url_resolver=url_resolver)

    def loader(self) -> Optional[LoaderFromRecordsDirectory]:
        return self._mysql_loader

    def loader_from_fileobj(self) -> Optional[LoaderFromFileobj]:
        return None

    def unloader(self) -> None:
        return None

    # https://dev.mysql.com/doc/refman/8.0/en/integer-types.html
    def integer_limits(self,
                       type_: sqlalchemy.types.Integer) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.dialects.mysql.TINYINT):
            if type_.unsigned:
                return IntegerType.UINT8.range
            else:
                return IntegerType.INT8.range
        elif isinstance(type_, sqlalchemy.dialects.mysql.SMALLINT):
            if type_.unsigned:
                return IntegerType.UINT16.range
            else:
                return IntegerType.INT16.range
        elif isinstance(type_, sqlalchemy.dialects.mysql.MEDIUMINT):
            if type_.unsigned:
                return IntegerType.UINT24.range
            else:
                return IntegerType.INT24.range
        elif isinstance(type_, sqlalchemy.dialects.mysql.INTEGER):
            if type_.unsigned:
                return IntegerType.UINT32.range
            else:
                return IntegerType.INT32.range
        elif isinstance(type_, sqlalchemy.dialects.mysql.BIGINT):
            if type_.unsigned:
                return IntegerType.UINT64.range
            else:
                return IntegerType.INT64.range
        return super().integer_limits(type_)

    def fp_constraints(self,
                       type_: sqlalchemy.types.Float) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.dialects.mysql.DOUBLE):
            return (64, FLOAT64_SIGNIFICAND_BITS)
        elif isinstance(type_, sqlalchemy.sql.sqltypes.FLOAT):
            return (32, FLOAT32_SIGNIFICAND_BITS)
        return super().fp_constraints(type_)

    def type_for_integer(self,
                         min_value: Optional[int],
                         max_value: Optional[int]) -> sqlalchemy.types.TypeEngine:
        """Find correct integral column type to fit the given min and max integer values"""

        if min_value is not None and max_value is not None:
            int_type = IntegerType.smallest_cover_for(min_value, max_value)
            if int_type == IntegerType.INT8:
                return sqlalchemy.dialects.mysql.TINYINT()
            elif int_type == IntegerType.UINT8:
                return sqlalchemy.dialects.mysql.TINYINT(unsigned=True)
            elif int_type == IntegerType.INT16:
                return sqlalchemy.sql.sqltypes.SMALLINT()
            elif int_type == IntegerType.UINT16:
                return sqlalchemy.dialects.mysql.SMALLINT(unsigned=True)
            elif int_type == IntegerType.INT24:
                return sqlalchemy.dialects.mysql.MEDIUMINT()
            elif int_type == IntegerType.UINT24:
                return sqlalchemy.dialects.mysql.MEDIUMINT(unsigned=True)
            elif int_type == IntegerType.INT32:
                return sqlalchemy.sql.sqltypes.INTEGER()
            elif int_type == IntegerType.UINT32:
                return sqlalchemy.dialects.mysql.INTEGER(unsigned=True)
            elif int_type == IntegerType.INT64:
                return sqlalchemy.sql.sqltypes.BIGINT()
            elif int_type == IntegerType.UINT64:
                return sqlalchemy.dialects.mysql.BIGINT(unsigned=True)
            else:
                num_digits_min = num_digits(min_value)
                num_digits_max = num_digits(max_value)
                digit_count = max(num_digits_min, num_digits_max)
                return self.type_for_fixed_point(precision=digit_count,
                                                 scale=0)
        return super().type_for_integer(min_value, max_value)

    def type_for_floating_point(self,
                                fp_total_bits: int,
                                fp_significand_bits: int) -> sqlalchemy.sql.sqltypes.Numeric:
        # https://dev.mysql.com/doc/refman/8.0/en/floating-point-types.html
        #
        # "A precision from 0 to 23 results in a 4-byte
        # single-precision FLOAT column. A precision from 24 to 53
        # results in an 8-byte double-precision DOUBLE column."
        if fp_significand_bits > FLOAT64_SIGNIFICAND_BITS:
            logger.warning(f"Falling back to MySQL DOUBLE type, as MySQL "
                           "doesn't support fp_significand_bits>{FLOAT64_SIGNIFICAND_BITS} "
                           f"(requested: {fp_significand_bits}")
            return sqlalchemy.sql.sqltypes.Float(precision=FLOAT64_SIGNIFICAND_BITS)
        return super().type_for_floating_point(fp_total_bits=fp_total_bits,
                                               fp_significand_bits=fp_significand_bits)

    def type_for_fixed_point(self,
                             precision: int,
                             scale: int) -> sqlalchemy.sql.sqltypes.Numeric:
        # "The maximum number of digits for DECIMAL is 65, but the
        # actual range for a given DECIMAL column can be constrained
        # by the precision or scale for a given column. When such a
        # column is assigned a value with more digits following the
        # decimal point than are permitted by the specified scale, the
        # value is converted to that scale. (The precise behavior is
        # operating system-specific, but generally the effect is
        # truncation to the permissible number of digits.)"
        #
        # https://dev.mysql.com/doc/refman/8.0/en/fixed-point-types.html
        if precision > 65:
            logger.warning('Using MySQL DOUBLE type to represent '
                           f'NUMERIC({precision},{scale}))')
            return sqlalchemy.dialects.mysql.DOUBLE()
        else:
            return super().type_for_fixed_point(precision=precision,
                                                scale=scale)

    def varchar_length_is_in_chars(self) -> bool:
        # This is assuming folks are using MySQL 5+
        # https://stackoverflow.com/questions/1997540/mysql-varchar-lengths-and-utf-8
        return True

    def type_for_date_plus_time(self, has_tz: bool = False) -> sqlalchemy.sql.sqltypes.DateTime:
        # Support six digits of fractional seconds to match other
        # databases and general expectations for a datetime
        #
        # Never has timezone, as the one type with a timezone
        # (TIMESTAMP) doesn't allow for dates before Jan 1, 1970, so
        # it's not generally useful.
        return sqlalchemy.dialects.mysql.DATETIME(fsp=6)



import io
from typing import IO, List


class ConcatFiles(io.RawIOBase):
    _files: List[IO[bytes]]

    def __init__(self, files: List[IO[bytes]]) -> None:
        self._files = files
        self._tell = 0

    def close(self) -> None:
        for f in self._files:
            f.close()
        self._files = []
        return super().close()

    def readable(self) -> bool:
        return True

    def readall(self) -> bytes:
        # Not sure if Records Mover has any uses of this which may be
        # done on large files, but bytearray() concatentation is
        # suggested by:
        #
        # https://www.guyrutenberg.com/2020/04/04/fast-bytes-concatenation-in-python/
        out = bytearray()
        while self._files:
            f = self._files.pop(0)
            chunk = f.read()
            self._tell += len(chunk)
            out = out + chunk
        return out

    def tell(self) -> int:
        return self._tell

    def read(self, size: int = -1) -> bytes:
        #  "When size is omitted or negative, the entire contents of
        #  the file will be read and returned"
        #
        # https://docs.python.org/3/tutorial/inputoutput.html
        if size < 0:
            return self.readall()

        while len(self._files) > 0:
            chunk = self._files[0].read(size)
            # If we aren't getting any bytes from this stream, lets move on to the next stream
            if len(chunk) == 0:
                f = self._files.pop(0)
                f.close()
            else:
                self._tell += len(chunk)
                return chunk

        return b''



from .sources import (RecordsSource, SupportsMoveToRecordsDirectory,
                      SupportsToFileobjsSource,
                      FileobjsSource, SupportsToDataframesSource)
from .targets.base import (RecordsTarget, SupportsMoveFromRecordsDirectory,
                           MightSupportMoveFromTempLocAfterFillingIt,
                           MightSupportMoveFromFileobjsSource,
                           SupportsMoveFromDataframes)
from .sources import base as sources_base
from .targets import base as targets_base
from .processing_instructions import ProcessingInstructions
from .records_format import BaseRecordsFormat
from .results import MoveResult
import logging

logger = logging.getLogger(__name__)


def move(records_source: RecordsSource,
         records_target: RecordsTarget,
         processing_instructions: ProcessingInstructions = ProcessingInstructions()) -> MoveResult:
    """Copy records from one location to another.  Applies a sequence of
    possible techniques to do this in an efficient way and respects
    the preferences set in records_source, records_target and
    processing_instructions.


    Example use:

    .. code-block:: python

       records = session.records
       db_engine = session.get_default_db_engine()
       url = 's3://some-bucket/some-directory/'
       source = records.sources.directory_from_url(url=url)
       target = records.targets.table(schema_name='myschema',
                                      table_name='mytable',
                                      db_engine=db_engine)
       results = records.move(source, target)

    :param records_source: object returned by a factory method in
       :class:`records_mover.records.sources.RecordsSources` which represents the place we're
       copying records from.
    :param records_target: object returned by a factory method in
       :class:`records_mover.records.targets.RecordsTargets` which represents the place we're
       copying records to.
    :param processing_instructions: Directives on how to handle different situations when
       processing files.
    :type processing_instructions: records_mover.records.ProcessingInstructions

    :rtype: records_mover.records.MoveResult
    """
    records_source.validate()
    records_target.validate()
    # This method works by looking for whether copy-related methods
    # are implemented on the source and target.

    # To figure out how to get your new source or target to work well
    # with move, follow this function down from the top and implement
    # the first method which makes sense for your new source/target.
    # See documentation for RecordsSource in sources.py and
    # RecordsTarget in targets.py for the semantics of the methods
    # being called.
    if (isinstance(records_source, sources_base.SupportsRecordsDirectory) and
        isinstance(records_target, SupportsMoveFromRecordsDirectory) and
       records_target.
        can_move_directly_from_scheme(records_source.records_directory().loc.scheme) and
       records_target.can_move_from_format(records_source.records_format)):
        # Tell the destination to load directly from wherever the
        # source is, without needing to make any copies of the data or
        # streaming it through the current box.
        directory = records_source.records_directory()
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    f"by moving from records directory in {directory}...")
        return records_target.\
            move_from_records_directory(directory=directory,
                                        override_records_format=records_source.records_format,
                                        processing_instructions=processing_instructions)
    elif (isinstance(records_source, FileobjsSource) and
          isinstance(records_target, MightSupportMoveFromFileobjsSource) and
          records_target.can_move_from_fileobjs_source() and
          records_target.can_move_from_format(records_source.records_format)):
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    "by moving directly from stream...")
        # See if we can stream from the source to the destination directly
        return records_target.move_from_fileobjs_source(records_source,
                                                        processing_instructions)
    elif (isinstance(records_source, SupportsMoveToRecordsDirectory) and
          isinstance(records_target, targets_base.SupportsRecordsDirectory) and
          records_source.can_move_to_scheme(records_target.records_directory().loc.scheme) and
          records_source.has_compatible_format(records_target)):
        # if target can accept records and doesn't specify a
        # records_format, or uses the same as the source, we can just
        # dump bytes directly!
        directory = records_target.records_directory()
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    f"by writing to {directory.scheme} records directory")
        pi = processing_instructions
        records_format = records_source.compatible_format(records_target)
        assert records_format is not None  # we checked compatibility above
        records_target.pre_load_hook()
        out = records_source.\
            move_to_records_directory(records_directory=directory,
                                      records_format=records_format,
                                      processing_instructions=pi)
        records_target.post_load_hook(num_rows_loaded=out.move_count)
        return out
    elif (isinstance(records_source, sources_base.SupportsRecordsDirectory) and
          isinstance(records_target, SupportsMoveFromRecordsDirectory) and
          records_target.can_move_from_format(records_source.records_format)):
        directory = records_source.records_directory()
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    "by loading from {directory.scheme} directory...")
        return records_target.\
            move_from_records_directory(directory=directory,
                                        override_records_format=records_source.records_format,
                                        processing_instructions=processing_instructions)
    elif isinstance(records_source, SupportsToFileobjsSource):
        target_records_format: BaseRecordsFormat = getattr(records_target, "records_format", None)
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    f"by first writing {records_source} to {target_records_format} "
                    "records format (if easy to rewrite)...")
        with records_source.\
                to_fileobjs_source(processing_instructions=processing_instructions,
                                   records_format_if_possible=target_records_format)\
                as fileobjs_source:
            return move(fileobjs_source, records_target, processing_instructions)
    elif (isinstance(records_source, SupportsMoveToRecordsDirectory) and
          isinstance(records_target, MightSupportMoveFromTempLocAfterFillingIt) and
          records_source.has_compatible_format(records_target) and
          records_source.
            can_move_to_scheme(records_target.temporary_loadable_directory_scheme()) and
          records_target.can_move_from_temp_loc_after_filling_it()):
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    f"by filling in a temporary location...")
        return records_target.move_from_temp_loc_after_filling_it(records_source,
                                                                  processing_instructions)
    elif (isinstance(records_source, SupportsToDataframesSource) and
          isinstance(records_target, SupportsMoveFromDataframes)):
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    f"by converting to dataframe...")
        with records_source.to_dataframes_source(processing_instructions) as dataframes_source:
            return records_target.\
                move_from_dataframes_source(dfs_source=dataframes_source,
                                            processing_instructions=processing_instructions)
    elif (isinstance(records_source, SupportsToDataframesSource)):
        logger.info(f"Mover: copying from {records_source} to {records_target} "
                    f"by first converting to dataframe...")
        with records_source.to_dataframes_source(processing_instructions) as dataframes_source:
            return move(dataframes_source, records_target, processing_instructions)
    else:
        raise NotImplementedError(f"Please teach me how to move records between "
                                  f"{records_source} and {records_target}")



import pandas as pd
from pandas import Series, Index
from typing import Any, Type, TYPE_CHECKING, Optional, Mapping, Union
from .statistics import RecordsSchemaFieldStringStatistics
from ...processing_instructions import ProcessingInstructions
from .representation import RecordsSchemaFieldRepresentation
from ....utils.limits import IntegerType
from .numpy import details_from_numpy_dtype
import numpy as np
if TYPE_CHECKING:
    from ..field import RecordsSchemaField  # noqa
    from ..schema import RecordsSchema  # noqa
    from pandas.core.dtypes.dtypes import ExtensionDtype # noqa

# Cribbed from non-public https://github.com/pandas-dev/pandas/blob/v1.2.1/pandas/_typing.py
Dtype = Union[
    "ExtensionDtype", str, np.dtype, Type[Union[str, float, int, complex, bool, object]]
]
DtypeObj = Union[np.dtype, "ExtensionDtype"]


def supports_nullable_ints() -> bool:
    """Detects if this version of pandas supports nullable int extension types."""
    return 'Int64Dtype' in dir(pd)


def integer_type_mapping(use_extension_types: bool) -> Mapping[IntegerType, DtypeObj]:
    if use_extension_types:
        return {
            IntegerType.INT8: pd.Int8Dtype(),
            IntegerType.UINT8: pd.UInt8Dtype(),
            IntegerType.INT16: pd.Int16Dtype(),
            IntegerType.UINT16: pd.UInt16Dtype(),
            IntegerType.INT24: pd.Int32Dtype(),
            IntegerType.UINT24: pd.Int32Dtype(),
            IntegerType.INT32: pd.Int32Dtype(),
            IntegerType.UINT32: pd.UInt32Dtype(),
            IntegerType.INT64: pd.Int64Dtype(),
            IntegerType.UINT64: pd.UInt64Dtype(),
        }
    else:
        return {
            IntegerType.INT8: np.int8,
            IntegerType.UINT8: np.uint8,
            IntegerType.INT16: np.int16,
            IntegerType.UINT16: np.uint16,
            IntegerType.INT24: np.int32,
            IntegerType.UINT24: np.uint32,
            IntegerType.INT32: np.int32,
            IntegerType.UINT32: np.uint32,
            IntegerType.INT64: np.int64,
            IntegerType.UINT64: np.uint64,
        }


def integer_type_for_range(min_: int, max_: int, has_extension_types: bool) -> Optional[DtypeObj]:
    int_type = IntegerType.smallest_cover_for(min_, max_)
    if int_type:
        return integer_type_mapping(has_extension_types).get(int_type)
    else:
        return None


def field_from_index(index: Index,
                     processing_instructions: ProcessingInstructions) -> 'RecordsSchemaField':

    from ..field import RecordsSchemaField  # noqa
    field_type, constraints = details_from_numpy_dtype(index.dtype, unique=True)
    representations = {
        'origin': RecordsSchemaFieldRepresentation.from_index(index)
    }

    return RecordsSchemaField(name=index.name,
                              field_type=field_type,
                              constraints=constraints,
                              statistics=None,  # call refine_from_dataframe() for stats
                              representations=representations)


def field_from_series(series: Series,
                      processing_instructions: ProcessingInstructions) -> 'RecordsSchemaField':
    from ..field import RecordsSchemaField  # noqa
    field_type, constraints = details_from_numpy_dtype(series.dtype, unique=False)
    representations = {
        'origin': RecordsSchemaFieldRepresentation.from_series(series)
    }

    return RecordsSchemaField(name=series.name,
                              field_type=field_type,
                              constraints=constraints,
                              statistics=None,  # call refine_from_dataframe() for stats
                              representations=representations)


def refine_field_from_series(field: 'RecordsSchemaField',
                             series: Series,
                             total_rows: int,
                             rows_sampled: int) -> 'RecordsSchemaField':
    from ..field import RecordsSchemaField  # noqa
    #
    # if the series is full of object types that aren't numpy
    # types that show up directly as `.dtype` already, we can find
    # that out.
    #
    unique_python_types = series.map(type).unique()
    if unique_python_types.size == 1:
        unique_python_type: Type[Any] = unique_python_types[0]
        field_type = field.python_type_to_field_type(unique_python_type)
        if field_type is not None:
            if RecordsSchemaField.is_more_specific_type(field_type, field.field_type):
                field = field.cast(field_type)

    if field.field_type == 'string':
        max_column_length = series.astype('str').map(len).max()
        if not np.isnan(max_column_length):
            statistics =\
                RecordsSchemaFieldStringStatistics(rows_sampled=rows_sampled,
                                                   total_rows=total_rows,
                                                   max_length_bytes=None,
                                                   max_length_chars=max_column_length)
            if field.statistics is None:
                field.statistics = statistics
            elif not isinstance(field.statistics, RecordsSchemaFieldStringStatistics):
                raise ValueError("Did not expect to see existing statistics "
                                 f"for string type: {field.statistics}")
            else:
                field.statistics.merge(statistics)
    return field



import sys
import types


__all__ = [
    '__version__',
    'Session',
    'Records',
    'records',
    'set_stream_logging',
    'move',
]

from . import records
from .version import __version__
from .session import Session
from .records import Records, move
from .logging import set_stream_logging
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .records.sources import RecordsSources
    from .records.targets import RecordsTargets


#
# The following magic allows for importing 'sources' and 'targets'
# with some sensible defaults.
#
# Note: there's a better way to do this once Python 3.6 is EOL
# (2021-12-23)
#
# https://stackoverflow.com/questions/1462986/lazy-module-variables-can-it-be-done
class _Sneaky(types.ModuleType):
    @property
    def _session(self) -> Session:
        if not hasattr(self, '_session_'):
            self._session_ = Session()
            self._session_.set_stream_logging()
        return self._session_

    @property
    def _records(self) -> Records:
        if not hasattr(self, '_records_'):
            self._records_ = self._session.records
        return self._records_

    @property
    def sources(self) -> 'RecordsSources':
        if not hasattr(self, '_sources'):
            self._sources = self._records.sources
        return self._sources

    @property
    def targets(self) -> 'RecordsTargets':
        if not hasattr(self, '_targets'):
            self._targets = self._records.targets
        return self._targets


sys.modules[__name__].__class__ = _Sneaky



from enum import Enum


class ExistingTableHandling(Enum):
    """Specifies behavior when an existing table with the same name is
    found when loading data into a database.
    """

    DELETE_AND_OVERWRITE = 1
    """Delete data transactionally (typically with a SQL DELETE statement)
    and then add new data to the existing table.  The delete and the
    load will be done in a single transaction if the database allows
    for that, but note that some do not, especially while using the
    most efficient load method.

    """

    TRUNCATE_AND_OVERWRITE = 2
    """Remove data from the existing table without regard for
    transactions, data integrity constraints, triggers or redo logs,
    typically using a SQL TRUNCATE statement.  The specific method
    depends on the database type.  This is typically the fastest way
    to clear a table, but please read your database documentation
    first and understand the consequences.
    """

    DROP_AND_RECREATE = 3
    """Remove the target table entirely, typically with a SQL DROP TABLE
    command."""

    APPEND = 4
    """Leave the target table and current data in place, and add data to
    table.  Note that Records Mover uses the most efficient method of
    loading data into the table, which may not honor triggers and
    integrity constraints.
    """





def canonicalize_db_type(db_type: str) -> str:
    canonical_db_type = {
        'vsql': 'vertica',
        'postgresql': 'postgres',
        'psql (redshift)': 'redshift',
        'psql': 'postgres',
        'Redshift': 'redshift'
    }
    lowercased_type = db_type.lower()
    return canonical_db_type.get(lowercased_type,
                                 lowercased_type)


def db_protocol(db_type: str) -> str:
    final_protocol = {
        'redshift': 'postgres',
    }
    return final_protocol.get(db_type, db_type)



from .types import HintCompression, HintEncoding, HintDateFormat, HintTimeOnlyFormat, HintQuoting
import logging
import csv
from typing import Optional, Dict


logger = logging.getLogger(__name__)

pandas_compression_from_hint: Dict[HintCompression, Optional[str]] = {
    'GZIP': 'gzip',
    'BZIP': 'bz2',
    None: None,
}

pandas_quoting_from_hint: Dict[HintQuoting, int] = {
    'minimal': csv.QUOTE_MINIMAL,
    'all': csv.QUOTE_ALL,
    'nonnumeric': csv.QUOTE_NONNUMERIC,
    None: csv.QUOTE_NONE
}

python_encoding_from_hint: Dict[Optional[HintEncoding], str] = {
    # valid python names: https://docs.python.org/3/library/codecs.html#standard-encodings
    # valid hints names: https://github.com/bluelabsio/knowledge/blob/master/Engineering/
    #    Architecture/JobDataExchange/output-design.md
    'UTF8': 'utf-8',
    'UTF16': 'utf-16',
    'UTF16LE': 'utf-16-le',
    'UTF16BE': 'utf-16-be',
    'LATIN1': 'latin_1',
    'CP1252': 'cp1252',
    'UTF8BOM': 'utf-8-sig',
    # Python will auto-detect UTF-16 with a BOM, but not UTF-8
    'UTF16BOM': 'utf-16',
}

# This is a Union because the date formats currently don't really
# account for MM/DD time.
#
# https://github.com/bluelabsio/records-mover/issues/75
python_date_format_from_hints: Dict[HintDateFormat, str] = {
    'DD-MM-YYYY': '%d-%m-%Y',
    'MM-DD-YYYY': '%m-%d-%Y',
    'YYYY-MM-DD': '%Y-%m-%d',
    'MM/DD/YY': '%m/%d/%y',
    'DD/MM/YY': '%d/%m/%y',
    'DD-MM-YY': '%d-%m-%y',
}

python_time_format_from_hints: Dict[HintTimeOnlyFormat, str] = {
    'HH24:MI:SS': '%H:%M:%S',
    'HH:MI:SS': '%H:%M:%S',
    'HH12:MI AM': '%I:%M %p',
}

hint_encoding_from_pandas: Dict[str, HintEncoding] = {
    'utf-8': 'UTF8',
    'utf-16': 'UTF16',
    'utf-16-le': 'UTF16LE',
    'utf-16-be': 'UTF16BE',
}

hint_encoding_from_chardet: Dict[str, HintEncoding] = {
    'UTF-8-SIG': 'UTF8BOM',
    'UTF-16': 'UTF16',
    'ISO-8859-1': 'LATIN1',
    # For some reason this is lowercase:
    #  https://github.com/chardet/chardet/blob/17218468eb16b7d0068bce7e4d20bac70f0bf555/chardet/utf8prober.py#L51
    'utf-8': 'UTF8',
    # But let's be ready if they change their minds:
    'UTF-8': 'UTF8',
    'Windows-1252': 'CP1252',
    # even if the only data it saw was in ASCII, let's be ready to see more
    'ascii': 'UTF8',
}



import logging
import json
import sqlalchemy as sa
from urllib.parse import quote_plus
from records_mover.creds.lpass import db_facts_from_lpass
from records_mover.db.db_type import canonicalize_db_type
from db_facts.db_facts_types import DBFacts
from typing import Union


logger = logging.getLogger(__name__)

# cheap translator between configured types (e.g., from
# cred-service/LastPass/etc) into which driver to use.
db_driver_for_type = {
    'postgres': 'postgresql',
    # pymysql is pure Python and is known to work correctly with LOAD
    # DATA LOCAL INFILE in SQLAlchemy, which mysqlclient did not as of
    # 2020-04.
    'mysql': 'mysql+pymysql',
    # vertica_python has the advantage of being pure Python and an
    # offering directly from Vertica:
    # https://github.com/vertica/vertica-python
    'vertica': 'vertica+vertica_python',
}

odbc_driver_for_type = {
    'vertica': 'vertica+pyodbc',
}

query_for_type = {
    'mysql': {
        # Please see SECURITY.md for security implications!
        "local_infile": True
    },
}


def create_vertica_odbc_sqlalchemy_url(db_facts: DBFacts) -> str:
    # Vertica wants the port in its ODBC connect string as a separate
    # parameter called "Port":
    #
    # https://my.vertica.com/docs/7.1.x/HTML/Content/Authoring/
    #   ConnectingToHPVertica/ClientODBC/DSNParameters.htm
    #
    # Unfortunately, sqlalchemy wants to provide port as part of the
    # 'Server' argument:
    #
    # https://github.com/zzzeek/sqlalchemy/blob/master/lib/sqlalchemy/
    #    connectors/pyodbc.py#L84
    #
    # As a result, we can't use the standard sqlalchemy URL format.

    db_url = ("Driver=HPVertica;"
              "SERVER={host};"
              "DATABASE={database};"
              "PORT={port};"
              "UID={user};"
              "PWD={password};"
              "CHARSET=UTF8;").\
              format(host=db_facts['host'],
                     database=db_facts['database'],
                     port=db_facts['port'],
                     user=db_facts['user'],
                     password=db_facts['password'])
    db_url = quote_plus(db_url)
    return "vertica+pyodbc:///?odbc_connect={}".format(db_url)


def create_bigquery_sqlalchemy_url(db_facts: DBFacts) -> str:
    "Create URL compatible with https://github.com/mxmzdlv/pybigquery"

    default_project_id = db_facts.get('bq_default_project_id')
    default_dataset_id = db_facts.get('bq_default_dataset_id')
    url = 'bigquery://'
    if default_project_id is not None:
        url += default_project_id
        if default_dataset_id is not None:
            url += '/'
            url += default_dataset_id
    return url


def create_bigquery_db_engine(db_facts: DBFacts) -> sa.engine.Engine:
    service_account_json = db_facts.get('bq_service_account_json')
    credentials_info = None
    if service_account_json is not None:
        credentials_info = json.loads(service_account_json)
        logger.info(f"Logging into BigQuery as {credentials_info['client_email']}")
    else:
        logger.info("Found no service account info for BigQuery, using local creds")
    url = create_bigquery_sqlalchemy_url(db_facts)
    return sa.engine.create_engine(url, credentials_info=credentials_info)


def create_sqlalchemy_url(db_facts: DBFacts,
                          prefer_odbc: bool = False) -> Union[str, sa.engine.url.URL]:
    db_type = canonicalize_db_type(db_facts['type'])
    driver = db_driver_for_type.get(db_type, db_type)
    if prefer_odbc:
        driver = odbc_driver_for_type.get(db_type, driver)
    # 'user' is the preferred key, but handle legacy code as well
    # still using 'username'
    username = db_facts.get('username', db_facts.get('user'))  # type: ignore
    if driver == 'vertica+pyodbc':
        return create_vertica_odbc_sqlalchemy_url(db_facts)
    elif driver == 'bigquery':
        if 'bq_service_account_json' in db_facts:
            raise NotImplementedError("pybigquery does not support providing credentials info "
                                      "(service account JSON) directly")

        return create_bigquery_sqlalchemy_url(db_facts)
    else:
        return sa.engine.url.URL(drivername=driver,
                                 username=username,
                                 password=db_facts['password'],
                                 host=db_facts['host'],
                                 port=db_facts['port'],
                                 database=db_facts['database'],
                                 query=query_for_type.get(db_type))


def engine_from_lpass_entry(lpass_entry_name: str) -> sa.engine.Engine:
    db_facts = db_facts_from_lpass(lpass_entry_name)
    return engine_from_db_facts(db_facts)


def engine_from_db_facts(db_facts: DBFacts) -> sa.engine.Engine:
    db_type = canonicalize_db_type(db_facts['type'])
    driver = db_driver_for_type.get(db_type, db_type)
    if driver == 'bigquery':
        # without writing creds to a temp file, pybigquery doesn't
        # support specifying service account creds in a URL - so let's
        # use create_engine() instead of creating a URL just in case.
        return create_bigquery_db_engine(db_facts)
    else:
        db_url = create_sqlalchemy_url(db_facts)
        return sa.create_engine(db_url)



from .constraints.string import RecordsSchemaFieldStringConstraints
from .statistics import RecordsSchemaFieldStringStatistics
from typing import Optional, Union, TYPE_CHECKING
if TYPE_CHECKING:
    from ....db import DBDriver  # noqa


def generate_string_length(constraints: Optional[RecordsSchemaFieldStringConstraints],
                           statistics: Optional[RecordsSchemaFieldStringStatistics],
                           driver: Optional['DBDriver']) -> int:
    len_from_constraints = string_length_from_field_details(constraints, driver)
    len_from_statistics = string_length_from_field_details(statistics, driver)

    if len_from_constraints is not None:
        # trust this first, as it may have more intent from the
        # creator of the origin representation.
        return len_from_constraints

    if len_from_statistics is not None:
        return len_from_statistics

    # safe-ish default if we can't figure out any better.
    return 256


def string_length_from_field_details(details: Optional[Union[RecordsSchemaFieldStringConstraints,
                                                             RecordsSchemaFieldStringStatistics]],
                                     driver: Optional['DBDriver']) -> Optional[int]:
    if details is not None:
        if driver is None or not driver.varchar_length_is_in_chars():
            # the target tracks varchar length in bytes

            if details.max_length_bytes is not None:
                # Note: this doesn't yet take into account
                # difference in encodings; if source encoded in
                # utf-8 and was mostly plain ASCII and destination
                # is in UTF-16 or something, you're gonna have a
                # bad time.
                return details.max_length_bytes
            elif details.max_length_chars is not None:
                # Note: also not yet taking into account
                # variations in source vs destination encoding
                utf_8_max_tax = 4
                # if every character uses a maximum number of
                # bytes, we'll need this much space...
                return details.max_length_chars * utf_8_max_tax
            else:
                pass  # use the default guess above
        else:
            # in chars
            if details.max_length_chars is not None:
                return details.max_length_chars
            elif details.max_length_bytes is not None:
                return details.max_length_bytes
            else:
                pass  # don't try to guess
    return None



from sqlalchemy.sql.expression import text, TextClause
from records_mover.utils import quiet_remove
from records_mover.records.delimited import cant_handle_hint
from records_mover.records import DelimitedRecordsFormat
from records_mover.records.delimited import (
    HintEncoding, HintRecordTerminator,
    HintFieldDelimiter, HintQuoteChar,
    HintQuoting, HintEscape,
    HintHeaderRow, HintCompression,
    HintDoublequote,
)
from typing import Optional, Set, Dict, NamedTuple, NoReturn, TYPE_CHECKING
if TYPE_CHECKING:
    from typing_extensions import Literal
    # http://dev.mysql.com/doc/refman/8.0/en/charset-mysql.html
    MySqlCharacterSet = Literal['big5', 'binary', 'latin1', 'ucs2',
                                'utf8', 'utf8mb4', 'utf16', 'utf16le',
                                'utf32']
else:
    MySqlCharacterSet = str


# The NoReturn annotations are the mypy way of validating we're
# covering all officially approved cases of an the hint at
# typechecking time.
#
# https://github.com/python/mypy/issues/6366#issuecomment-560369716
def _assert_never(the_bad_hint: NoReturn) -> NoReturn:
    assert False


MYSQL_CHARACTER_SETS_FOR_LOAD: Dict[HintEncoding, MySqlCharacterSet] = {
    "UTF8": 'utf8',
    "UTF16": 'utf16',
    "UTF16LE": 'utf16le',
    # https://dev.mysql.com/doc/refman/8.0/en/charset-unicode.html
    #
    # The MySQL implementation of UCS-2, UTF-16, and UTF-32 stores
    # characters in big-endian byte order and does not use a byte
    # order mark (BOM) at the beginning of values. Other database
    # systems might use little-endian byte order or a BOM. In such
    # cases, conversion of values will need to be performed when
    # transferring data between those systems and MySQL. The
    # implementation of UTF-16LE is little-endian.
    "UTF16BE": 'utf16',
    # MySQL uses no BOM for UTF-8 values.
    "LATIN1": 'latin1',
    #  MySQL's latin1 is the same as the Windows cp1252 character set.
    #
    # https://dev.mysql.com/doc/refman/8.0/en/charset-we-sets.html
    "CP1252": 'latin1',
}


# Mark as total=False so we can create this incrementally
class MySqlLoadOptions(NamedTuple):
    character_set: MySqlCharacterSet
    fields_terminated_by: str  # default '\t'
    fields_enclosed_by: Optional[str]  # default None
    fields_optionally_enclosed_by: Optional[str]  # default None
    fields_escaped_by: Optional[str]  # default '\'
    lines_starting_by: str  # default ''
    lines_terminated_by: str  # default '\n'
    ignore_n_lines: int

    def generate_load_data_sql(self,
                               filename: str,
                               schema_name: str,
                               table_name: str) -> TextClause:
        sql = f"""\
LOAD DATA
LOCAL INFILE :filename
INTO TABLE {schema_name}.{table_name}
CHARACTER SET :character_set
FIELDS
    TERMINATED BY :fields_terminated_by
"""
        if self.fields_enclosed_by is not None and self.fields_optionally_enclosed_by is None:
            sql += "    ENCLOSED BY :fields_enclosed_by\n"
        elif self.fields_enclosed_by is None and self.fields_optionally_enclosed_by is not None:
            sql += "    OPTIONALLY ENCLOSED BY :fields_optionally_enclosed_by\n"
        elif self.fields_enclosed_by is not None and self.fields_optionally_enclosed_by is not None:
            raise SyntaxError('fields_enclosed_by and optionally_fields_enclosed_by '
                              'cannot both be set')
        if self.fields_escaped_by is not None:
            sql += "    ESCAPED BY :fields_escaped_by\n"
        sql += """\
LINES
    STARTING BY :lines_starting_by
    TERMINATED BY :lines_terminated_by
IGNORE :ignore_n_lines LINES
"""
        clause = text(sql)
        clause = clause.\
            bindparams(filename=filename,
                       character_set=self.character_set,
                       fields_terminated_by=self.fields_terminated_by,
                       lines_starting_by=self.lines_starting_by,
                       lines_terminated_by=self.lines_terminated_by,
                       ignore_n_lines=self.ignore_n_lines)
        if self.fields_enclosed_by is not None:
            clause = clause.bindparams(fields_enclosed_by=self.
                                       fields_enclosed_by)
        if self.fields_optionally_enclosed_by is not None:
            clause = clause.bindparams(fields_optionally_enclosed_by=self.
                                       fields_optionally_enclosed_by)
        if self.fields_escaped_by is not None:
            clause = clause.bindparams(fields_escaped_by=self.
                                       fields_escaped_by)
        return clause


def mysql_load_options(unhandled_hints: Set[str],
                       records_format: DelimitedRecordsFormat,
                       fail_if_cant_handle_hint: bool) -> MySqlLoadOptions:
    hints = records_format.validate(fail_if_cant_handle_hint=fail_if_cant_handle_hint)

    #
    # The server uses the character set indicated by the
    # character_set_database system variable to interpret the
    # information in the file. SET NAMES and the setting of
    # character_set_client do not affect interpretation of input. If
    # the contents of the input file use a character set that differs
    # from the default, it is usually preferable to specify the
    # character set of the file by using the CHARACTER SET clause. A
    # character set of binary specifies “no conversion.”
    #
    hint_encoding: HintEncoding = hints.encoding
    character_set = MYSQL_CHARACTER_SETS_FOR_LOAD.get(hint_encoding)
    if character_set is not None:
        mysql_character_set = character_set
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'encoding', records_format.hints)
        mysql_character_set = 'utf8'
    quiet_remove(unhandled_hints, 'encoding')

    field_terminator: HintFieldDelimiter = hints.field_delimiter
    mysql_fields_terminator = field_terminator
    quiet_remove(unhandled_hints, 'field-delimiter')

    # https://dev.mysql.com/doc/refman/8.0/en/load-data.html
    #
    #
    # LOAD DATA can be used to read files obtained from external
    # sources. For example, many programs can export data in
    # comma-separated values (CSV) format, such that lines have fields
    # separated by commas and enclosed within double quotation marks,
    # with an initial line of column names. If the lines in such a
    # file are terminated by carriage return/newline pairs, the
    # statement shown here illustrates the field- and line-handling
    # options you would use to load the file:
    #
    # LOAD DATA INFILE 'data.txt' INTO TABLE tbl_name
    #  FIELDS TERMINATED BY ',' ENCLOSED BY '"'
    #  LINES TERMINATED BY '\r\n'
    #  IGNORE 1 LINES;
    #
    #
    mysql_fields_enclosed_by = None
    mysql_fields_optionally_enclosed_by = None
    hint_quotechar: HintQuoteChar = hints.quotechar
    hint_quoting: HintQuoting = hints.quoting
    if hint_quoting == 'all':
        mysql_fields_enclosed_by = hint_quotechar
    elif hint_quoting == 'minimal':
        # "If the input values are not necessarily enclosed within
        # quotation marks, use OPTIONALLY before the ENCLOSED BY option."
        #
        # This implies to me that parsing here is permissive -
        # otherwise unambiguous strings without double quotes around
        # them will be understood as a string, not rejected.
        mysql_fields_optionally_enclosed_by = hint_quotechar
    elif hint_quoting == 'nonnumeric':
        mysql_fields_optionally_enclosed_by = hint_quotechar
    elif hint_quoting is None:
        pass
    else:
        _assert_never(hint_quotechar)
    quiet_remove(unhandled_hints, 'quoting')
    quiet_remove(unhandled_hints, 'quotechar')

    # If the field begins with the ENCLOSED BY character, instances of
    # that character are recognized as terminating a field value only
    # if followed by the field or line TERMINATED BY sequence. To
    # avoid ambiguity, occurrences of the ENCLOSED BY character within
    # a field value can be doubled and are interpreted as a single
    # instance of the character. For example, if ENCLOSED BY '"' is
    # specified, quotation marks are handled as shown here:
    hint_doublequote: HintDoublequote = hints.doublequote
    if hint_quoting is not None:
        # We need to ignore flake8's "is vs ==" check because 'is'
        # doesn't work currently with MyPy's Literal[] case checking
        if hint_doublequote == True:  # noqa: E712
            pass
        elif hint_doublequote == False:  # noqa: E712
            cant_handle_hint(fail_if_cant_handle_hint, 'doublequote', records_format.hints)
        else:
            _assert_never(hint_doublequote)
    quiet_remove(unhandled_hints, 'doublequote')

    # FIELDS ESCAPED BY controls how to read or write special characters:
    #
    # * For input, if the FIELDS ESCAPED BY character is not empty,
    #   occurrences of that character are stripped and the following
    #   character is taken literally as part of a field value. Some
    #   two-character sequences that are exceptions, where the first
    #   character is the escape character.
    #
    # [...]
    #
    # If the FIELDS ESCAPED BY character is empty, escape-sequence
    # interpretation does not occur.
    hint_escape: HintEscape = hints.escape
    if hint_escape is None:
        mysql_fields_escaped_by = None
    elif hint_escape == '\\':
        mysql_fields_escaped_by = '\\'
    else:
        _assert_never(hint_quoting)
    quiet_remove(unhandled_hints, 'escape')

    mysql_lines_starting_by = ''

    hint_record_terminator: HintRecordTerminator = hints.record_terminator
    mysql_lines_terminated_by = hint_record_terminator
    quiet_remove(unhandled_hints, 'record-terminator')

    hint_header_row: HintHeaderRow = hints.header_row
    # We need to ignore flake8's "is vs ==" check because 'is'
    # doesn't work currently with MyPy's Literal[] case checking
    if hint_header_row == True:  # noqa: E712
        mysql_ignore_n_lines = 1
    elif hint_header_row == False:  # noqa: E712
        mysql_ignore_n_lines = 0
    else:
        _assert_never(hint_header_row)
    quiet_remove(unhandled_hints, 'header-row')

    hint_compression: HintCompression = hints.compression
    if hint_compression is not None:
        cant_handle_hint(fail_if_cant_handle_hint, 'compression', records_format.hints)
    quiet_remove(unhandled_hints, 'compression')

    #
    # Testing of date/time parsing in MySQL has shown it to be pretty
    # conservative.
    #
    # That said, for DD/MM and MM/DD support, we may want to look into
    # "set trade_date" per
    # https://stackoverflow.com/questions/44171283/load-data-local-infile-with-sqlalchemy-and-pymysql
    #
    if 'YYYY-MM-DD' not in hints.datetimeformat or 'AM' in hints.datetimeformat:
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformat', records_format.hints)
    quiet_remove(unhandled_hints, 'datetimeformat')
    if 'YYYY-MM-DD' not in hints.datetimeformattz or 'AM' in hints.datetimeformattz:
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformat', records_format.hints)
    quiet_remove(unhandled_hints, 'datetimeformattz')
    if hints.dateformat != 'YYYY-MM-DD':
        cant_handle_hint(fail_if_cant_handle_hint, 'dateformat', records_format.hints)
    quiet_remove(unhandled_hints, 'dateformat')
    if 'AM' in hints.timeonlyformat:
        cant_handle_hint(fail_if_cant_handle_hint, 'timeonlyformat', records_format.hints)
    quiet_remove(unhandled_hints, 'timeonlyformat')

    return MySqlLoadOptions(character_set=mysql_character_set,
                            fields_terminated_by=mysql_fields_terminator,
                            fields_enclosed_by=mysql_fields_enclosed_by,
                            fields_optionally_enclosed_by=mysql_fields_optionally_enclosed_by,
                            fields_escaped_by=mysql_fields_escaped_by,
                            lines_starting_by=mysql_lines_starting_by,
                            lines_terminated_by=mysql_lines_terminated_by,
                            ignore_n_lines=mysql_ignore_n_lines)



from records_mover.records.schema import RecordsSchema
from records_mover.records.records_format import BaseRecordsFormat
from records_mover.db import DBDriver
from records_mover.records.prep import TablePrep
from records_mover.records.table import TargetTableDetails
from records_mover.records.processing_instructions import ProcessingInstructions
import logging

logger = logging.getLogger(__name__)


class BaseTableMoveAlgorithm:
    def __init__(self,
                 prep: TablePrep,
                 target_table_details: TargetTableDetails,
                 processing_instructions: ProcessingInstructions) -> None:
        self.prep = prep
        self.tbl = target_table_details
        self.processing_instructions = processing_instructions

    def schema_sql_for_load(self,
                            records_schema: RecordsSchema,
                            records_format: BaseRecordsFormat,
                            driver: 'DBDriver') -> str:
        tweaked_records_schema = driver.tweak_records_schema_for_load(records_schema,
                                                                      records_format)
        return tweaked_records_schema.to_schema_sql(driver,
                                                    self.tbl.schema_name,
                                                    self.tbl.table_name)



import inspect
from ..mover_types import JsonSchema
from typing import Optional, List, Union, TypeVar


# default is set to inspect.Parameter.empty if no default
DefaultValue = object

A = TypeVar('A')


class JsonSchemaDocument:
    """Represents JSON Schema"""
    def __init__(self,
                 json_type: Union[str, List[str]],
                 default: DefaultValue = inspect.Parameter.empty,
                 enum: Optional[List[A]] = None,
                 description: Optional[str] = None) -> None:
        self.json_type = json_type
        self.default = default
        self.enum = enum
        self.description = description

    def to_data(self) -> JsonSchema:
        out: JsonSchema = {
            'type': self.json_type
        }
        if self.enum is not None:
            out['enum'] = self.enum
        if self.description is not None:
            out['description'] = self.description
        if not self.default == inspect.Parameter.empty:
            out['default'] = self.default
        return out



import datetime
from ...processing_instructions import ProcessingInstructions
import logging
from typing import Optional, Dict, Any, Type, cast, TYPE_CHECKING
from ....utils.limits import (FLOAT16_SIGNIFICAND_BITS,
                              FLOAT32_SIGNIFICAND_BITS,
                              FLOAT64_SIGNIFICAND_BITS,
                              FLOAT80_SIGNIFICAND_BITS)
from .representation import RecordsSchemaFieldRepresentation
from .constraints import (RecordsSchemaFieldConstraints,
                          RecordsSchemaFieldIntegerConstraints,
                          RecordsSchemaFieldDecimalConstraints)
from .statistics import RecordsSchemaFieldStatistics
from .field_types import RECORDS_FIELD_TYPES
if TYPE_CHECKING:
    from pandas import Series, Index
    from sqlalchemy import Column
    from sqlalchemy.types import TypeEngine
    from records_mover.db import DBDriver  # noqa
    from .field_types import FieldType
    from .pandas import Dtype

    from mypy_extensions import TypedDict

    class MandatoryFieldDict(TypedDict):
        type: FieldType

    from .statistics import FieldStatisticsDict
    from .constraints import FieldConstraintsDict
    from .representation import FieldRepresentationDict

    class FieldDict(MandatoryFieldDict, total=False):
        index: int
        constraints: FieldConstraintsDict
        statistics: FieldStatisticsDict
        representations: Dict[str, FieldRepresentationDict]

logger = logging.getLogger(__name__)

DEFAULT_COLUMN_LENGTH = 256


class RecordsSchemaField:
    def __init__(self,
                 name,
                 field_type,
                 constraints,
                 statistics,
                 representations):
        # type: (str, FieldType, Optional[RecordsSchemaFieldConstraints], Optional[RecordsSchemaFieldStatistics], Dict[str, RecordsSchemaFieldRepresentation]) -> None  # noqa
        self.name = name
        self.field_type = field_type
        self.constraints = constraints
        self.statistics = statistics
        self.representations = representations

    def refine_from_series(self,
                           series: 'Series',
                           total_rows: int,
                           rows_sampled: int) -> 'RecordsSchemaField':
        from .pandas import refine_field_from_series
        return refine_field_from_series(self, series, total_rows, rows_sampled)

    @staticmethod
    def is_more_specific_type(a: 'FieldType', b: 'FieldType') -> bool:
        if b == 'string' and a != 'string':
            return True
        return False

    @staticmethod
    def python_type_to_field_type(specific_type: Type[Any]) -> Optional['FieldType']:
        import numpy as np
        import pandas as pd

        # Note: records spec doesn't cover complex number types, so
        # np.complex_, complex64 and complex128 are not supported
        # except as a string.
        type_mapping: Dict[Type[Any], 'FieldType'] = {
            # https://docs.scipy.org/doc/numpy-1.14.0/user/basics.types.html
            int: 'integer',
            np.int64: 'integer',
            np.int_: 'integer',
            np.intc: 'integer',
            np.intp: 'integer',
            np.int8: 'integer',
            np.int16: 'integer',
            np.int32: 'integer',
            np.int64: 'integer',
            np.uint8: 'integer',
            np.uint16: 'integer',
            np.uint32: 'integer',
            np.uint64: 'integer',

            np.float_: 'decimal',
            np.float16: 'decimal',
            np.float32: 'decimal',
            np.float64: 'decimal',
            np.float128: 'decimal',
            float: 'decimal',

            str: 'string',
            np.str_: 'string',
            np.object_: 'string',

            bool: 'boolean',
            np.bool_: 'boolean',

            datetime.date: 'date',

            datetime.time: 'time',

            datetime.datetime: 'datetime',

            pd.Timestamp: 'datetime',
        }
        if specific_type not in type_mapping:
            logger.warning(f"Please teach me how to map {specific_type} into records "
                           "schema field types")
        return type_mapping.get(specific_type)

    @staticmethod
    def from_index(index: 'Index',
                   processing_instructions: ProcessingInstructions) -> 'RecordsSchemaField':
        from .pandas import field_from_index
        return field_from_index(index=index,
                                processing_instructions=processing_instructions)

    @staticmethod
    def from_series(series: 'Series',
                    processing_instructions: ProcessingInstructions) -> 'RecordsSchemaField':
        from .pandas import field_from_series
        return field_from_series(series=series,
                                 processing_instructions=processing_instructions)

    @staticmethod
    def from_sqlalchemy_column(column: 'Column',
                               driver: 'DBDriver',
                               rep_type: str)\
            -> 'RecordsSchemaField':
        from .sqlalchemy import field_from_sqlalchemy_column

        return field_from_sqlalchemy_column(column=column,
                                            driver=driver,
                                            rep_type=rep_type)

    def to_sqlalchemy_type(self,
                           driver: 'DBDriver') -> 'TypeEngine':
        from .sqlalchemy import field_to_sqlalchemy_type

        return field_to_sqlalchemy_type(self, driver)

    def to_sqlalchemy_column(self, driver: 'DBDriver') -> 'Column':
        from .sqlalchemy import field_to_sqlalchemy_column

        return field_to_sqlalchemy_column(self, driver)

    def cast_series_type(self, series: 'Series') -> 'Series':
        import pandas as pd
        if self.field_type == 'time':
            if series.size > 0:
                # https://stackoverflow.com/questions/34501930/how-to-convert-timedelta-to-time-of-day-in-pandas
                #
                # Some databases (e.g., MySQL) contains a TIME type
                # which is ambiguous - it can either represent a
                # particular time of day or it can represent an
                # elapsed amount of time.
                #
                # Clever, right?
                #
                # Unfortunately, Pandas knows about time deltas, but
                # not about times of day, so upon use of read_sql(),
                # these objects will come out as as a timedelta64[ns]
                # type.
                #
                # Since that's not what our 'time' field type means,
                # we have to convert it back to a string, or when it
                # gets turned into a CSV later, it'll look really
                # goofy - 1pm will come out as: "0 days 01:00:00".
                #
                if type(series[0]) == pd.Timedelta:
                    # https://stackoverflow.com/questions/34501930/how-to-convert-timedelta-to-time-of-day-in-pandas

                    # Convert from "0 days 12:34:56.000000000" to "12:34:56"
                    def components_to_time_str(df: pd.DataFrame) -> datetime.time:
                        return datetime.time(hour=df['hours'],
                                             minute=df['minutes'],
                                             second=df['seconds'])
                    logger.debug("Applying pd.Timedelta logic on series for %s", self.name)
                    out = series.dt.components.apply(axis=1, func=components_to_time_str)
                    return out

        target_type = self.to_pandas_dtype()
        logger.debug("Casting field %s from type %r to type %s", self.name, series.dtype,
                     target_type)
        return series.astype(target_type)

    def to_pandas_dtype(self) -> 'Dtype':
        import numpy as np
        import pandas as pd
        from .pandas import supports_nullable_ints, integer_type_for_range

        has_extension_types = supports_nullable_ints()

        if self.field_type == 'integer':
            int_constraints =\
                cast(Optional[RecordsSchemaFieldIntegerConstraints], self.constraints)
            min_: Optional[int] = None
            max_: Optional[int] = None
            required = False
            if int_constraints:
                min_ = int_constraints.min_
                max_ = int_constraints.max_
                required = int_constraints.required

            if not required and not has_extension_types:
                logger.warning(f"Dataframe field {self.name} is nullable, but using pandas "
                               f"{pd.__version__} which does not support nullable integer type")

            if min_ is not None and max_ is not None:
                dtype = integer_type_for_range(min_, max_, has_extension_types)
                if dtype:
                    return dtype
                else:
                    logger.warning("Asked for a type larger than int64 in dataframe "
                                   f"field '{self.name}' - providing float128, but "
                                   "loss of precision will occur!  "
                                   f"Requested min/max values: {min_}/{max_}")
                    return np.float128
            else:
                logger.warning(f"No integer constraints provided for field '{self.name}'; "
                               "using int64")
                if has_extension_types:
                    return pd.Int64Dtype()
                else:
                    return np.int64
        elif self.field_type == 'decimal':
            decimal_constraints =\
                cast(Optional[RecordsSchemaFieldDecimalConstraints], self.constraints)
            if decimal_constraints:
                if (decimal_constraints.fixed_precision is not None and
                   decimal_constraints.fixed_scale is not None):
                    logger.warning("Pandas doesn't support a fixed precision type - "
                                   "using np.float64")
                    return np.float64
                elif (decimal_constraints.fp_total_bits is not None and
                      decimal_constraints.fp_significand_bits is not None):
                    if (decimal_constraints.fp_total_bits <= 16 and
                       decimal_constraints.fp_significand_bits <= FLOAT16_SIGNIFICAND_BITS):
                        return np.float16
                    elif (decimal_constraints.fp_total_bits <= 32 and
                          decimal_constraints.fp_significand_bits <= FLOAT32_SIGNIFICAND_BITS):
                        return np.float32
                    elif (decimal_constraints.fp_total_bits <= 64 and
                          decimal_constraints.fp_significand_bits <= FLOAT64_SIGNIFICAND_BITS):
                        return np.float64
                    elif (decimal_constraints.fp_total_bits <= 80 and
                          decimal_constraints.fp_significand_bits <= FLOAT80_SIGNIFICAND_BITS):
                        return np.float128
                    else:
                        logger.warning("Downgrading float type to np.float128.  "
                                       "Requested total bits: "
                                       f"{decimal_constraints.fp_total_bits}.  "
                                       "Requested significand bits: "
                                       f"{decimal_constraints.fp_significand_bits}")
                        return np.float128

            logger.warning(f"No decimal constraints provided for field '{self.name}'; "
                           "using float64")
            return np.float64
        elif self.field_type == 'boolean':
            return np.bool_
        elif self.field_type == 'string':
            return np.object_
        elif self.field_type == 'date':
            return np.object_
        elif self.field_type == 'datetime':
            return 'datetime64[ns]'
        elif self.field_type == 'datetimetz':
            return 'datetime64[ns, UTC]'
        elif self.field_type == 'time':
            return np.object_
        elif self.field_type == 'timetz':
            return np.object_
        else:
            raise NotImplementedError("Teach me how to handle records schema "
                                      f"type {self.field_type}")

    def to_data(self) -> 'FieldDict':
        out: 'FieldDict' = {'type': self.field_type}
        if self.constraints:
            out['constraints'] = self.constraints.to_data()
        if self.statistics:
            out['statistics'] = self.statistics.to_data()
        if self.representations:
            out['representations'] = {
                system: rep.to_data()
                for system, rep in self.representations.items()
            }
        return out

    @staticmethod
    def from_data(name: str, data: 'FieldDict') -> 'RecordsSchemaField':
        representations = {}
        field_type = data['type']
        if field_type not in RECORDS_FIELD_TYPES:
            logger.warning(f"Unknown field_type ({field_type})--consider upgrading records-mover")
        for rep_name, rep_data in data.get('representations', {}).items():
            field_representation = RecordsSchemaFieldRepresentation.from_data(rep_data)
            if field_representation is not None:
                representations[rep_name] = field_representation
        return RecordsSchemaField(name=name,
                                  field_type=data['type'],
                                  constraints=RecordsSchemaFieldConstraints.
                                  from_data(data.get('constraints'), field_type=field_type),
                                  statistics=RecordsSchemaFieldStatistics.
                                  from_data(data.get('statistics'), field_type=field_type),
                                  representations=representations)

    def cast(self, field_type: 'FieldType') -> 'RecordsSchemaField':
        if self.field_type == field_type:
            return self
        if self.constraints is None:
            constraints = None
        else:
            constraints = self.constraints.cast(field_type)
        if self.statistics is None:
            statistics = None
        else:
            statistics = self.statistics.cast(field_type)
        field = RecordsSchemaField(name=self.name,
                                   field_type=field_type,
                                   constraints=constraints,
                                   statistics=statistics,
                                   representations=self.representations)
        return field



from records_mover.utils import quiet_remove
from records_mover.records.delimited import cant_handle_hint, ValidatedRecordsHints
from typing import Optional, Set
from .types import DateOrderStyle


def determine_input_date_order_style(unhandled_hints: Set[str],
                                     hints: ValidatedRecordsHints,
                                     fail_if_cant_handle_hint: bool) ->\
        Optional[DateOrderStyle]:
    date_order_style: Optional[DateOrderStyle] = None

    def upgrade_date_order_style(style: DateOrderStyle, hint_name: str) -> None:
        nonlocal date_order_style
        if date_order_style not in (None, style):
            cant_handle_hint(fail_if_cant_handle_hint, hint_name, hints)
        else:
            date_order_style = style
            quiet_remove(unhandled_hints, hint_name)

    # https://www.postgresql.org/docs/9.5/datatype-datetime.html#DATATYPE-DATETIME-INPUT

    # "Date and time input is accepted in almost any reasonable
    # format, including ISO 8601, SQL-compatible, traditional
    # POSTGRES, and others. For some formats, ordering of day, month,
    # and year in date input is ambiguous and there is support for
    # specifying the expected ordering of these fields. Set the
    # DateStyle parameter to MDY to select month-day-year
    # interpretation, DMY to select day-month-year interpretation, or
    # YMD to select year-month-day interpretation."

    # $ cd tests/integration
    # $ ./itest shell
    # $ db dockerized-postgres
    # postgres=# SHOW DateStyle;
    #  DateStyle
    # -----------
    #  ISO, MDY
    # (1 row)
    #
    # postgres=#

    datetimeformattz = hints.datetimeformattz

    # datetimeformattz: Valid values: "YYYY-MM-DD HH:MI:SSOF",
    # "YYYY-MM-DD HH:MI:SS", "YYYY-MM-DD HH24:MI:SSOF", "YYYY-MM-DD
    # HH24:MI:SSOF", "MM/DD/YY HH24:MI". See Redshift docs for more
    # information (note that HH: is equivalent to HH24: and that if
    # you don't provide an offset (OF), times are assumed to be in
    # UTC).

    # Default value is "YYYY-MM-DD HH:MI:SSOF".

    #
    # To get a postgres database to test these cases with:
    #
    # cd tests/integration
    # ./itest shell
    # db dockerized-postgres
    if datetimeformattz in ['YYYY-MM-DD HH:MI:SSOF',
                            'YYYY-MM-DD HH24:MI:SSOF']:
        #
        #  postgres=# select timestamptz '2020-01-01 23:01:01-10';
        #        timestamptz
        #  ------------------------
        #   2020-01-02 09:01:01+00
        #  (1 row)
        #
        #  postgres=#

        # Any DateStyle will do as this is unambiguous
        quiet_remove(unhandled_hints, 'datetimeformattz')
    elif datetimeformattz == 'YYYY-MM-DD HH:MI:SS':
        #
        #
        #  postgres=# select timestamptz '2020-01-01 23:01:01';
        #        timestamptz
        #  ------------------------
        #   2020-01-01 23:01:01+00
        #  (1 row)
        #
        #  postgres=#

        # Any DateStyle will do as this is unambiguous
        quiet_remove(unhandled_hints, 'datetimeformattz')
    elif datetimeformattz == "MM/DD/YY HH24:MI":
        # "MM/DD/YY HH24:MI"
        #
        #  postgres=# select timestamptz '01/02/2999 23:01';
        #        timestamptz
        #  ------------------------
        #   2999-01-02 23:01:00+00
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('MDY', 'datetimeformattz')
    elif datetimeformattz == 'DD-MM-YY HH:MI:SSOF':
        #  postgres=# select timestamptz '02-01-2999 23:01:01-10';
        #        timestamptz
        #  ------------------------
        #   2999-02-02 09:01:01+00
        #  (1 row)
        #
        upgrade_date_order_style('DMY', 'datetimeformattz')
    elif datetimeformattz == 'DD/MM/YY HH:MI:SSOF':
        #  postgres=# select timestamptz '02/01/99 23:01:01-10';
        #        timestamptz
        #  ------------------------
        #   1999-02-02 09:01:01+00
        #  (1 row)
        #
        upgrade_date_order_style('DMY', 'datetimeformattz')
    elif datetimeformattz == 'DD-MM-YYYY HH:MI:SSOF':
        #  postgres=# select timestamptz '02-01-1999 23:01:01-10';
        #        timestamptz
        #  ------------------------
        #   1999-02-02 09:01:01+00
        #  (1 row)
        #
        upgrade_date_order_style('DMY', 'datetimeformattz')
    elif datetimeformattz == 'MM/DD/YY HH:MI:SSOF':
        #  postgres=# select timestamptz '01/02/99 23:01:01-10';
        #        timestamptz
        #  ------------------------
        #   1999-01-03 09:01:01+00
        #  (1 row)
        #
        upgrade_date_order_style('MDY', 'datetimeformattz')
    elif datetimeformattz == 'MM-DD-YYYY HH:MI:SSOF':
        #  postgres=# select timestamptz '01-02-1999 23:01:01-10';
        #        timestamptz
        #  ------------------------
        #   1999-01-03 09:01:01+00
        #  (1 row)
        #
        upgrade_date_order_style('MDY', 'datetimeformattz')
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformattz', hints)

    # datetimeformat: Valid values: "YYYY-MM-DD HH24:MI:SS",
    # "YYYY-MM-DD HH12:MI AM", "MM/DD/YY HH24:MI". See Redshift docs
    # for more information.

    datetimeformat = hints.datetimeformat

    if datetimeformat in ("YYYY-MM-DD HH24:MI:SS",
                          "YYYY-MM-DD HH:MI:SS"):
        #
        #  postgres=# select timestamp '2020-01-02 15:13:12';
        #        timestamp
        #  ---------------------
        #   2020-01-02 15:13:12
        #  (1 row)
        #
        #  postgres=#

        # Any DateStyle will do as this is unambiguous
        quiet_remove(unhandled_hints, 'datetimeformat')
    elif datetimeformat == "YYYY-MM-DD HH12:MI AM":
        # "YYYY-MM-DD HH12:MI AM"
        #
        #  postgres=# select timestamp '2020-01-02 1:13 PM';
        #        timestamp
        #  ---------------------
        #   2020-01-02 13:13:00
        #  (1 row)
        #
        #  postgres=#

        # Any DateStyle will do as this is unambiguous
        quiet_remove(unhandled_hints, 'datetimeformat')
    elif datetimeformat == "MM/DD/YY HH24:MI":
        #  postgres=# select timestamp '01/02/20 15:23';
        #        timestamp
        #  ---------------------
        #   2020-01-02 15:23:00
        #  (1 row)
        #
        #  postgres=#

        upgrade_date_order_style('MDY', 'datetimeformat')
    elif datetimeformat == 'DD-MM-YY HH12:MI AM':
        #  postgres=# select timestamp '02-01-20 3:23 PM';
        #        timestamp
        #  ---------------------
        #   2020-02-01 15:23:00
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('DMY', 'datetimeformat')
    elif datetimeformat == 'DD/MM/YY HH12:MI AM':
        #  postgres=# select timestamp '02/01/20 3:23 PM';
        #        timestamp
        #  ---------------------
        #   2020-02-01 15:23:00
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('DMY', 'datetimeformat')
    elif datetimeformat == 'DD-MM-YYYY HH12:MI AM':
        #  postgres=# select timestamp '02-01-2020 3:23 PM';
        #        timestamp
        #  ---------------------
        #   2020-02-01 15:23:00
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('DMY', 'datetimeformat')
    elif datetimeformat == 'MM-DD-YY HH12:MI AM':
        #  postgres=# select timestamp '02-01-20 3:23 PM';
        #        timestamp
        #  ---------------------
        #   2020-02-01 15:23:00
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('MDY', 'datetimeformat')
    elif datetimeformat == 'MM/DD/YY HH12:MI AM':
        #  postgres=# select timestamp '02/01/20 3:23 PM';
        #        timestamp
        #  ---------------------
        #   2020-02-01 15:23:00
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('MDY', 'datetimeformat')
    elif datetimeformat == 'MM-DD-YYYY HH12:MI AM':
        #  postgres=# select timestamp '01-02-2020 3:23 PM';
        #        timestamp
        #  ---------------------
        #   2020-01-02 15:23:00
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('MDY', 'datetimeformat')
    elif datetimeformat == 'MM-DD-YYYY HH:MI:SS':
        #  postgres=# select timestamp '01-02-2020 3:23 PM';
        #        timestamp
        #  ---------------------
        #   2020-01-02 15:23:00
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('MDY', 'datetimeformat')
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'datetimeformat', hints)

    timeonlyformat = hints.timeonlyformat

    # timeonlyformat: Valid values: "HH12:MI AM" (e.g., "1:00 PM"),
    # "HH24:MI:SS" (e.g., "13:00:00")

    if timeonlyformat == "HH12:MI AM":
        # "HH12:MI AM" (e.g., "1:00 PM"),
        #
        #  postgres=# select time '1:00 PM';
        #     time
        #  ----------
        #   13:00:00
        #  (1 row)
        #
        #  postgres=#

        # Supported!
        quiet_remove(unhandled_hints, 'timeonlyformat')
    elif timeonlyformat == "HH24:MI:SS":

        # "HH24:MI:SS" (e.g., "13:00:00")
        #
        #  postgres=# select time '13:00:00';
        #     time
        #  ----------
        #   13:00:00
        #  (1 row)
        #
        #  postgres=#

        # Supported!
        quiet_remove(unhandled_hints, 'timeonlyformat')
    elif timeonlyformat == "HH:MI:SS":
        #  postgres=# select time '13:00:00';
        #     time
        #  ----------
        #   13:00:00
        #  (1 row)
        #
        #  postgres=#
        quiet_remove(unhandled_hints, 'timeonlyformat')
    elif timeonlyformat == "HH24:MI":

        # "HH24:MI" (e.g., "13:00")
        #
        #  postgres=# select time '13:00';
        #     time
        #  ----------
        #   13:00:00
        #  (1 row)
        #
        #  postgres=#

        # Supported!
        quiet_remove(unhandled_hints, 'timeonlyformat')
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'timeonlyformat', hints)

    # dateformat: Valid values: null, "YYYY-MM-DD", "MM-DD-YYYY", "DD-MM-YYYY", "MM/DD/YY".
    dateformat = hints.dateformat

    if dateformat == "YYYY-MM-DD":
        #  postgres=# select date '1999-01-02';
        #      date
        #  ------------
        #   1999-01-02
        #  (1 row)
        #
        #  postgres=#
        # Any DateStyle will do as this is unambiguous
        quiet_remove(unhandled_hints, 'dateformat')
    elif dateformat == "MM-DD-YYYY":
        # "MM-DD-YYYY"
        #
        #  postgres=# select date '01-02-1999';
        #      date
        #  ------------
        #   1999-01-02
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('MDY', 'dateformat')
    elif dateformat == "DD-MM-YYYY":
        # "DD-MM-YYYY" - not supported by default, need to switch to
        # DMY:
        #
        #  postgres=# select date '02-01-1999';
        #      date
        #  ------------
        #   1999-02-01
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('DMY', 'dateformat')
    elif dateformat == "DD-MM-YY":
        #  postgres=# select date '02-01-99';
        #      date
        #  ------------
        #   1999-02-01
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('DMY', 'dateformat')
    elif dateformat == "MM/DD/YY":
        # "MM/DD/YY".
        #
        #  postgres=# select date '01/02/99';
        #      date
        #  ------------
        #   1999-01-02
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('MDY', 'dateformat')
    elif dateformat == "DD/MM/YY":
        #  postgres=# select date '02/01/99';
        #      date
        #  ------------
        #   1999-02-01
        #  (1 row)
        #
        #  postgres=#
        upgrade_date_order_style('DMY', 'dateformat')
    elif dateformat is None:
        # null implies that that date format is unknown, and that the
        # implementation SHOULD generate using their default value and
        # parse permissively.

        # ...which is what Postgres does!
        quiet_remove(unhandled_hints, 'dateformat')
    else:
        cant_handle_hint(fail_if_cant_handle_hint, 'dateformat', hints)

    return date_order_style



from typing_extensions import Literal
from typing_inspect import get_args
from typing import List

FieldType = Literal['integer',
                    'decimal',
                    'string',
                    'boolean',
                    'date',
                    'time',
                    'timetz',
                    'datetime',
                    'datetimetz']

RECORDS_FIELD_TYPES: List[str] = list(get_args(FieldType))



import os
from db_facts.db_facts_types import DBFacts
from typing import Dict, Optional


def db_facts_from_env() -> DBFacts:
    db_facts: Dict[str, Optional[str]]
    if os.environ.get('DB_TYPE') == 'bigquery':
        db_facts = {
            'type': os.environ['DB_TYPE']
        }
        if 'BQ_DEFAULT_PROJECT_ID' in os.environ:
            db_facts['bq_default_project_id'] = os.environ['BQ_DEFAULT_PROJECT_ID']
        if 'BQ_DEFAULT_DATASET_ID' in os.environ:
            db_facts['bq_default_dataset_id'] = os.environ['BQ_DEFAULT_DATASET_ID']
        if 'BQ_SERVICE_ACCOUNT_JSON' in os.environ:
            db_facts['bq_service_account_json'] = os.environ['BQ_SERVICE_ACCOUNT_JSON']
    else:
        db_facts = {
            'host': os.environ.get('DB_HOST'),
            'database': os.environ.get('DB_DATABASE'),
            'port': os.environ.get('DB_PORT'),
            'user': os.environ.get('DB_USERNAME'),
            'password': os.environ.get('DB_PASSWORD'),
            'type': os.environ.get('DB_TYPE')
        }
        redshift_base_url_values = {
            name.lower(): value
            for name, value in os.environ.items()
            if name.startswith('REDSHIFT_SPECTRUM_BASE_URL_')
        }

        db_facts.update(redshift_base_url_values)

        if None in db_facts.values():
            raise NotImplementedError("Please run with with-db or set DB_* environment variables")
    return db_facts  # type: ignore



from contextlib import contextmanager
from sqlalchemy import MetaData
from ..records.unload_plan import RecordsUnloadPlan
from ..records.records_format import BaseRecordsFormat
from ..records.records_directory import RecordsDirectory
from records_mover.url.base import BaseDirectoryUrl
from typing import Union, List, Optional, Iterator
from abc import ABCMeta, abstractmethod
import sqlalchemy


class Unloader(metaclass=ABCMeta):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Connection, sqlalchemy.engine.Engine]) -> None:
        self.db = db
        self.meta = MetaData()

    @abstractmethod
    def unload(self,
               schema: str,
               table: str,
               unload_plan: RecordsUnloadPlan,
               directory: RecordsDirectory) -> Optional[int]:
        """Export data in the specified table to the RecordsDirectory instance
        named 'directory'.  Guarantees a manifest file named
        'manifest' is written to the target directory pointing to the
        target records.

        Returns number of rows loaded (if database provides that
        info)."""
        ...

    @abstractmethod
    def known_supported_records_formats_for_unload(self) -> List[BaseRecordsFormat]:
        """Supplies a list of the records formats which can be bulk exported
        from this database.  This may not be the full set - see
        can_unload_this_format() to test other possibilities.
        """
        ...

    @abstractmethod
    def can_unload_format(self, target_records_format: BaseRecordsFormat) -> bool:
        """Determine whether a specific records format can be exported by this
        database."""
        ...

    @abstractmethod
    def can_unload_to_scheme(self, scheme: str) -> bool:
        """Determine whether a URL scheme can be unloaded to by this database.
        Depending on capabilities of this driver and the underlying database,
        this may depend on whether a temporary location has been configured."""
        ...

    @abstractmethod
    @contextmanager
    def temporary_unloadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        """Provide a temporary directory which can be used for bulk export from
        this database and clean it up when done"""
        ...

    def best_records_format(self) -> Optional[BaseRecordsFormat]:
        """Returns the ideal records format for export by this database.
        Useful in the absence of other constraints."""
        supported_formats = self.known_supported_records_formats_for_unload()
        if len(supported_formats) == 0:
            return None
        return supported_formats[0]



import logging
from records_mover.mover_types import _assert_never
from typing import Optional, cast, TYPE_CHECKING
from records_mover.utils.limits import (FLOAT16_SIGNIFICAND_BITS,
                                        FLOAT32_SIGNIFICAND_BITS,
                                        FLOAT64_SIGNIFICAND_BITS,
                                        FLOAT80_SIGNIFICAND_BITS)
if TYPE_CHECKING:
    import sqlalchemy
    import numpy as np
    from records_mover.db import DBDriver  # noqa
    from mypy_extensions import TypedDict

    class MandatoryFieldConstraintsDict(TypedDict):
        required: bool

    class FieldConstraintsDict(MandatoryFieldConstraintsDict, total=False):
        unique: bool

    class FieldStringConstraintsDict(FieldConstraintsDict, total=False):
        max_length_bytes: int
        max_length_chars: int

    class FieldDecimalConstraintsDict(FieldConstraintsDict, total=False):
        fixed_precision: int
        fixed_scale: int
        fp_total_bits: int
        fp_significand_bits: int

    class FieldIntegerConstraintsDict(FieldConstraintsDict, total=False):
        min: str
        max: str

    from ..field_types import FieldType  # noqa


logger = logging.getLogger(__name__)


class RecordsSchemaFieldConstraints:
    def __init__(self, required: bool, unique: Optional[bool] = None):
        """
        :param required: If True, data must always be provided for this
        column in the origin representation; if False, a 'null' or
        equivalent optional value can exist instead.

        :param unique: If knowledge of a constraint is known, will be
        set to True if there is a uniqueness constraint on the field,
        False if there is not an explicit constraint on uniqueness.
        This is a constraint, so we're talking about whether a
        non-unique value *could* potentially be created in the origin
        representation, not whether one exists in practice.  If set to
        None, we have no knowledge of uniqueness constraints.
        """
        self.required = required
        self.unique = unique

    def to_data(self) -> 'FieldConstraintsDict':
        out: 'FieldConstraintsDict' = {
            'required': self.required,
        }
        if self.unique is not None:
            out['unique'] = self.unique

        return out

    @staticmethod
    def from_data(data: Optional['FieldConstraintsDict'],
                  field_type: 'FieldType') -> Optional['RecordsSchemaFieldConstraints']:
        from .string import RecordsSchemaFieldStringConstraints
        from .integer import RecordsSchemaFieldIntegerConstraints
        from .decimal import RecordsSchemaFieldDecimalConstraints
        if data is None:
            return None
        required = data.get('required', False)
        unique = data.get('unique')
        if field_type == 'string':
            string_data = cast('FieldStringConstraintsDict', data)
            return RecordsSchemaFieldStringConstraints(
                required=required,
                unique=unique,
                max_length_bytes=string_data.get('max_length_bytes'),
                max_length_chars=string_data.get('max_length_chars'))
        elif field_type == 'integer':
            int_data = cast('FieldIntegerConstraintsDict', data)
            min_str = int_data.get('min')
            max_str = int_data.get('max')
            min_: Optional[int] = None
            max_: Optional[int] = None
            if min_str:
                min_ = int(min_str)
            if max_str:
                max_ = int(max_str)
            return RecordsSchemaFieldIntegerConstraints(
                required=required,
                unique=unique,
                min_=min_,
                max_=max_)
        elif field_type == 'decimal':
            decimal_data = cast('FieldDecimalConstraintsDict', data)
            return RecordsSchemaFieldDecimalConstraints(
                required=required,
                unique=unique,
                fixed_precision=decimal_data.get("fixed_precision"),
                fixed_scale=decimal_data.get("fixed_scale"),
                fp_total_bits=decimal_data.get("fp_total_bits"),
                fp_significand_bits=decimal_data.get("fp_significand_bits"))
        else:
            return RecordsSchemaFieldConstraints(required=required, unique=unique)

    def cast(self, field_type: 'FieldType') -> 'RecordsSchemaFieldConstraints':
        from .integer import RecordsSchemaFieldIntegerConstraints
        from .decimal import RecordsSchemaFieldDecimalConstraints
        from .string import RecordsSchemaFieldStringConstraints
        required = self.required
        unique = self.unique
        constraints: RecordsSchemaFieldConstraints
        if field_type == 'integer':
            constraints =\
                RecordsSchemaFieldIntegerConstraints(required=required,
                                                     unique=unique,
                                                     min_=None,
                                                     max_=None)
        elif field_type == 'string':
            constraints =\
                RecordsSchemaFieldStringConstraints(required=required,
                                                    unique=unique,
                                                    max_length_bytes=None,
                                                    max_length_chars=None)
        elif field_type == 'decimal':
            constraints =\
                RecordsSchemaFieldDecimalConstraints(required=required,
                                                     unique=unique)
        elif (field_type == 'boolean' or
              field_type == 'date' or
              field_type == 'time' or
              field_type == 'timetz' or
              field_type == 'datetime' or
              field_type == 'datetimetz'):
            constraints =\
                RecordsSchemaFieldConstraints(required=required,
                                              unique=unique)
        else:
            _assert_never(field_type,
                          'Teach me how to downcast constraints '
                          f'for {field_type}')

        return constraints

    @staticmethod
    def from_sqlalchemy_type(required: bool,
                             unique: Optional[bool],
                             type_: 'sqlalchemy.types.TypeEngine',
                             driver: 'DBDriver') -> 'RecordsSchemaFieldConstraints':
        return RecordsSchemaFieldConstraints(required=required, unique=unique)

    @staticmethod
    def from_numpy_dtype(dtype: 'np.dtype',
                         unique: bool) -> 'RecordsSchemaFieldConstraints':
        from .string import RecordsSchemaFieldStringConstraints
        from .integer import RecordsSchemaFieldIntegerConstraints
        from .decimal import RecordsSchemaFieldDecimalConstraints

        SIGNED_INTEGER_KIND = 'i'
        UNSIGNED_INTEGER_KIND = 'u'
        FLOAT_KIND = 'f'
        if dtype.kind == SIGNED_INTEGER_KIND:
            bytes_length = dtype.itemsize
            # http://python-history.blogspot.com/2009/03/problem-with-integer-division.html
            max_ = 2 ** (bytes_length*8)//2-1
            min_ = 0 - 2 ** (bytes_length*8)//2
            return RecordsSchemaFieldIntegerConstraints(required=False, unique=unique,
                                                        min_=min_, max_=max_)
        elif dtype.kind == UNSIGNED_INTEGER_KIND:
            bytes_length = dtype.itemsize
            max_ = 2 ** (bytes_length*8) - 1
            min_ = 0
            return RecordsSchemaFieldIntegerConstraints(required=False, unique=unique,
                                                        min_=min_, max_=max_)
        elif dtype.kind == FLOAT_KIND:
            total_and_significand_precision = {
                2: (16, FLOAT16_SIGNIFICAND_BITS),
                4: (32, FLOAT32_SIGNIFICAND_BITS),
                8: (64, FLOAT64_SIGNIFICAND_BITS),
                16: (80, FLOAT80_SIGNIFICAND_BITS),
            }
            fp_total_bits: Optional[int] = None
            fp_significand_bits: Optional[int] = None
            if dtype.itemsize in total_and_significand_precision:
                fp_total_bits, fp_significand_bits = total_and_significand_precision[dtype.itemsize]
            else:
                logger.warning(f"Could not determine floating point constraints for {dtype}")
            return RecordsSchemaFieldDecimalConstraints(required=False,
                                                        unique=unique,
                                                        fp_total_bits=fp_total_bits,
                                                        fp_significand_bits=fp_significand_bits)
        elif dtype == 'O':
            return RecordsSchemaFieldStringConstraints(required=False,
                                                       unique=unique,
                                                       max_length_bytes=None,
                                                       max_length_chars=None)
        else:
            return RecordsSchemaFieldConstraints(required=False, unique=unique)

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.to_data()})"

    def __repr__(self) -> str:
        return self.__str__()



from .fileobjs import FileobjsSource
from urllib.parse import urlparse
from .base import SupportsToFileobjsSource
from contextlib import contextmanager
from ..delimited import sniff_compression_from_url
from ..processing_instructions import ProcessingInstructions
from ...url.resolver import UrlResolver
from ..records_format import BaseRecordsFormat
from ..schema import RecordsSchema
from .. import PartialRecordsHints
import logging
from typing import Optional, Iterator

logger = logging.getLogger(__name__)


class DataUrlRecordsSource(SupportsToFileobjsSource):
    def __init__(self,
                 input_url: str,
                 url_resolver: UrlResolver,
                 records_format: Optional[BaseRecordsFormat]=None,
                 records_schema: Optional[RecordsSchema]=None,
                 initial_hints: Optional[PartialRecordsHints]=None) -> None:
        self.input_url = input_url
        self.url_resolver = url_resolver
        self.records_format = records_format
        self.records_schema = records_schema
        self.initial_hints = initial_hints
        if self.initial_hints is None:
            self.initial_hints = {}
        if 'compression' not in self.initial_hints:
            # if we end up sniffing hints, we can start out assuming
            # compression type based on the filename.
            self.initial_hints['compression'] = sniff_compression_from_url(self.input_url)

    @contextmanager
    def to_fileobjs_source(self,
                           processing_instructions: ProcessingInstructions,
                           records_format_if_possible: Optional[BaseRecordsFormat]=None)\
            -> Iterator['FileobjsSource']:
        """Convert current source to a FileObjsSource and present it in a context manager"""
        with self.url_resolver.file_url(self.input_url).open() as fileobj:
            input_url_obj = urlparse(self.input_url)
            path = input_url_obj.path
            filename = path.split('/')[-1]
            with FileobjsSource.\
                infer_if_needed(target_names_to_input_fileobjs={filename: fileobj},
                                records_format=self.records_format,
                                records_schema=self.records_schema,
                                processing_instructions=processing_instructions,
                                initial_hints=self.initial_hints) as fileobjs_source:
                yield fileobjs_source

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.records_format})"



from .base_creds import BaseCreds
import logging
from db_facts.db_facts_types import DBFacts
from records_mover.logging import register_secret
from typing import Iterable, Optional, Union, TYPE_CHECKING
if TYPE_CHECKING:
    # see the 'gsheets' extras_require option in setup.py - needed for this!
    import google.auth.credentials  # noqa
    import boto3  # noqa


logger = logging.getLogger(__name__)


class CredsViaAirflow(BaseCreds):
    def boto3_session(self, aws_creds_name: str) -> 'boto3.session.Session':
        from airflow.contrib.hooks.aws_hook import AwsHook
        aws_hook = AwsHook(aws_creds_name)
        return aws_hook.get_session()

    def db_facts(self, db_creds_name: str) -> DBFacts:
        from airflow.hooks import BaseHook
        conn = BaseHook.get_connection(db_creds_name)
        out: DBFacts = {}

        def add(key: str, value: Optional[Union[str, int]]) -> None:
            if value is not None:
                out[key] = value  # type: ignore

        register_secret(conn.password)

        add('host', conn.host)
        add('port', conn.port)
        add('database', conn.schema)
        add('user', conn.login)
        add('password', conn.password)
        # conn.extra_dejson returns {} if no 'extra' is set in the Connection:
        # https://airflow.apache.org/docs/stable/_modules/airflow/models/connection.html
        add('type', conn.extra_dejson.get('type', conn.conn_type.lower()))
        add('bq_default_project_id', conn.extra_dejson.get('extra__google_cloud_platform__project'))
        add('bq_default_dataset_id', conn.extra_dejson.get('bq_default_dataset_id'))
        add('bq_service_account_json',
            conn.extra_dejson.get('extra__google_cloud_platform__keyfile_dict'))
        add('protocol', conn.extra_dejson.get('protocol'))

        return out

    def _gcp_creds(self, gcp_creds_name: str,
                   scopes: Iterable[str]) -> 'google.auth.credentials.Credentials':
        from records_mover.airflow.hooks.google_cloud_credentials_hook import (
            GoogleCloudCredentialsHook
        )
        gcp_hook = GoogleCloudCredentialsHook(gcp_conn_id=gcp_creds_name)
        for intended_scope in scopes:
            if intended_scope not in gcp_hook.scopes():
                logger.warning(f"{intended_scope} not configured as a scope in "
                               f"connection_id {gcp_creds_name}")
        return gcp_hook.get_conn()



"""Create and run jobs to convert between different sources and targets"""
from records_mover import Session
from ..processing_instructions import ProcessingInstructions
from ..results import MoveResult
from ...mover_types import JobConfig
from .config import config_to_args
import logging

logger = logging.getLogger(__name__)


def run_records_mover_job(source_method_name: str,
                          target_method_name: str,
                          job_name: str,
                          config: JobConfig) -> MoveResult:
    session = Session()
    try:
        source_method = getattr(session.records.sources, source_method_name)
        target_method = getattr(session.records.targets, target_method_name)
        logger.info('Starting...')

        source_kwargs = config_to_args(config=config['source'],
                                       method=source_method,
                                       session=session)
        target_kwargs = config_to_args(config=config['target'],
                                       method=target_method,
                                       session=session)
        pi_config = {k: config[k] for k in config if k not in ['source', 'target', 'func']}
        pi_kwargs = config_to_args(pi_config,
                                   method=ProcessingInstructions,
                                   session=session)
        processing_instructions = ProcessingInstructions(**pi_kwargs)

        records = session.records
        source = source_method(**source_kwargs)
        target = target_method(**target_kwargs)
        return records.move(source, target, processing_instructions)
    except Exception:
        logger.error('', exc_info=True)
        raise



from pandas import DataFrame

__all__ = [
    'pandas_to_csv_options',
    'pandas_read_csv_options',
    'prep_df_for_csv_output',
    'prep_df_for_loading',
]
from .to_csv_options import pandas_to_csv_options
from .read_csv_options import pandas_read_csv_options
from .prep_for_csv import prep_df_for_csv_output


def _lowercase_column_names(df: DataFrame) -> DataFrame:
    df.columns = map(str.lower, df.columns)
    return df


def prep_df_for_loading(df: DataFrame) -> DataFrame:
    return _lowercase_column_names(df)



from records_mover.mover_types import _ensure_all_cases_covered
from records_mover.records.records_types import RecordsFormatType
from .records_format import (
    BaseRecordsFormat, DelimitedRecordsFormat, ParquetRecordsFormat, AvroRecordsFormat
)
from ..url.base import BaseDirectoryUrl, BaseFileUrl
from .processing_instructions import ProcessingInstructions
from .delimited import PartialRecordsHints
from typing import cast
import logging

logger = logging.getLogger(__name__)


class RecordsFormatFile:
    def __init__(self, records_loc: BaseDirectoryUrl) -> None:
        self.records_loc = records_loc

    def load_format(self, fail_if_dont_understand: bool) -> BaseRecordsFormat:
        prefix = '_format_'
        matching_locs = self.records_loc.files_matching_prefix(prefix)
        if len(matching_locs) != 1:
            raise TypeError(f"_format file not found in bucket under {prefix}*")
        format_loc = matching_locs[0]
        format_type = format_loc.filename()[len(prefix):]
        # cast so we can ensure at build-time we have covered all
        # cases below
        format_type = cast(RecordsFormatType, format_type)
        if format_type == 'delimited':
            return self.load_delimited_format(format_loc, fail_if_dont_understand)
        elif format_type == 'parquet':
            return ParquetRecordsFormat()
        elif format_type == 'avro':
            return AvroRecordsFormat()
        else:
            _ensure_all_cases_covered(format_type)
            raise TypeError(f"Format type {format_type} not yet supported in this library")

    def load_delimited_format(self, format_loc: BaseFileUrl,
                              fail_if_dont_understand: bool) -> BaseRecordsFormat:
        data = format_loc.json_contents()
        variant = None
        hints: PartialRecordsHints = {}
        # For simple form (such as with avro or parquet), this file MUST be empty.
        if data is None:
            # For detailed form (such as delimited), it MUST contain
            # the same JSON as in the format field in the
            # configuration.
            raise TypeError(f"Detailed format information must be provided in {format_loc.url} "
                            f"for type delimited")
        if not isinstance(data, dict):
            raise TypeError(f"{format_loc.url} does not contain a JSON object as expected")
        if 'variant' not in data:
            raise TypeError(f"variant not specified in {format_loc.url}")
        potential_variant = data['variant']
        if not isinstance(potential_variant, str):
            raise TypeError(f"Invalid variant specified in {format_loc.url}: "
                            f"{potential_variant}")
        variant = potential_variant
        if 'hints' in data:
            hints = data['hints']
        processing_instructions =\
            ProcessingInstructions(fail_if_dont_understand=fail_if_dont_understand)
        return DelimitedRecordsFormat(variant=variant, hints=hints,
                                      processing_instructions=processing_instructions)

    def save_format(self, records_format: BaseRecordsFormat) -> None:
        file_loc = self.records_loc.file_in_this_directory(f"_format_{records_format.format_type}")
        contents = records_format.json()
        logger.info(f"Storing format info into {file_loc.url}")
        file_loc.store_string(contents)



"""CLI to move records from place to place"""
import argparse
from odictliteral import odict
from .job.schema import method_to_json_schema
from .job.mover import run_records_mover_job
from ..utils.json_schema import method_signature_to_json_schema
from .processing_instructions import ProcessingInstructions
from records_mover.cli.job_config_schema_as_args_parser import (JobConfigSchemaAsArgsParser,
                                                                arguments_output_to_config)
from records_mover.logging import set_stream_logging
from ..mover_types import JsonSchema, JobConfig
from ..version import __version__
import sys
from typing import Callable, Dict, Any, TYPE_CHECKING
if TYPE_CHECKING:
    from records_mover import Session


def populate_subparser(bootstrap_session: 'Session',
                       sub_parser: argparse.ArgumentParser,
                       source_method_name: str,
                       target_method_name: str,
                       subjob_name: str) -> JobConfig:
    source_method = getattr(bootstrap_session.records.sources, source_method_name)
    target_method = getattr(bootstrap_session.records.targets, target_method_name)
    job_config_schema = {
        "type": "object",
        "properties": odict[
            'source': method_to_json_schema(source_method),  # type: ignore
            'target': method_to_json_schema(target_method),  # type: ignore
        ],
        "required": ["source", "target"],
    }
    JobConfigSchemaAsArgsParser(config_json_schema=job_config_schema,
                                argument_parser=sub_parser).configure_arg_parser()
    return job_config_schema


def make_job_fn(source_method_name: str,
                target_method_name: str,
                name: str,
                job_config_schema: JsonSchema) -> Callable[[Dict[str, Any]], None]:
    def job_fn(raw_config: Dict[str, Any]) -> None:
        job_config = arguments_output_to_config(raw_config)
        run_records_mover_job(source_method_name,
                              target_method_name,
                              job_name=name,
                              config=job_config)
    return job_fn


def main() -> None:
    # https://github.com/googleapis/google-auth-library-python/issues/271
    import warnings
    warnings.filterwarnings("ignore",
                            "Your application has authenticated using end user credentials")

    # skip in-memory sources/targets like dataframes that don't make
    # sense from the command-line
    source_method_name_by_cli_name = {
        'table': 'table',
        'gsheet': 'google_sheet',
        'recordsdir': 'directory_from_url',
        'url': 'data_url',
        'file': 'local_file'
    }
    target_method_name_by_cli_name = {
        'gsheet': 'google_sheet',
        'table': 'table',
        'recordsdir': 'directory_from_url',
        'url': 'data_url',
        'file': 'local_file',
        'spectrum': 'spectrum',
    }
    sources = source_method_name_by_cli_name.keys()
    targets = target_method_name_by_cli_name.keys()

    description = 'Move tabular data ("records") from one place to another'
    parser = argparse.ArgumentParser(description=description,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    pi_config_schema =\
        method_signature_to_json_schema(ProcessingInstructions.__init__,
                                        special_handling={},
                                        parameters_to_ignore=['self'])
    JobConfigSchemaAsArgsParser(config_json_schema=pi_config_schema,
                                argument_parser=parser).configure_arg_parser()

    # https://stackoverflow.com/questions/15405636/pythons-argparse-to-show-programs-version-with-prog-and-version-string-formatt
    parser.add_argument('-V', '--version', action='version', version="%(prog)s ("+__version__+")")
    subparsers = parser.add_subparsers(help='subcommand_help')
    from records_mover import Session
    bootstrap_session = Session()

    for source in sources:
        for target in targets:
            name = f"{source}2{target}"
            sub_parser = subparsers.add_parser(name, help=f"Copy from {source} to {target}")
            source_method_name = source_method_name_by_cli_name[source]
            target_method_name = target_method_name_by_cli_name[target]
            job_config_schema = \
                populate_subparser(bootstrap_session,
                                   sub_parser, source_method_name, target_method_name,
                                   subjob_name=name)
            sub_parser.set_defaults(func=make_job_fn(source_method_name=source_method_name,
                                                     target_method_name=target_method_name,
                                                     name=name,
                                                     job_config_schema=job_config_schema))
    args = parser.parse_args()
    raw_config = vars(args)
    func = getattr(args, 'func', None)
    if func is None:
        parser.print_help()
    else:
        set_stream_logging()
        try:
            func(raw_config)
        except Exception:
            # This is logged above using a redacting logger
            sys.exit(1)



import json
import numpy as np
from pandas import DataFrame
from typing import Any


# http://stackoverflow.com/questions/27050108/convert-numpy-type-to-python
class NumPyJSONEncoder(json.JSONEncoder):
    def default(self, obj: object) -> object:
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumPyJSONEncoder, self).default(obj)


def json_dumps(item: str) -> Any:
    return json.dumps(item, cls=NumPyJSONEncoder)


def purge_unnamed_unused_columns(df: DataFrame) -> DataFrame:
    # Some spreadsheets will drop a trailing ',' after the
    # last column in the header with actual content when
    # exporting to CSV, leaving you with an empty column.
    # Pandas helpfully names this empty column for you as
    # "unnamed: 1", or maybe "Unnamed: 1" (not sure why/when
    # that differs).  Let's clean those up.
    for column in df:
        if column.startswith('Unnamed: ') or column.startswith('unnamed: '):
            if not df[column].notnull().any():
                df = df.drop(column, axis=1)
    return df



import io
from contextlib import contextmanager
from records_mover.records.delimited import PartialRecordsHints
from typing import Union, IO, Iterator, TYPE_CHECKING
from .conversions import (
    python_encoding_from_hint,
    pandas_compression_from_hint,
    pandas_quoting_from_hint
)
if TYPE_CHECKING:
    from pandas.io.parsers import TextFileReader  # noqa


@contextmanager
def stream_csv(filepath_or_buffer: Union[str, IO[bytes]],
               hints: PartialRecordsHints)\
               -> Iterator['TextFileReader']:
    """Returns a context manager that can be used to generate a full or
    partial dataframe from a CSV.  If partial, it will not read the
    entire CSV file into memory."""

    from pandas import read_csv

    header_row = hints.get('header-row')
    header: Union[str, int, None]
    if header_row is None:
        # we weren't told...
        header = 'infer'
    elif header_row:
        # there's a header row in the file, so tell Pandas that row 0
        # is the headers:
        header = 0
    else:
        # there's no header row in the file
        header = None
    compression_hint = hints.get('compression')
    encoding_hint = hints.get('encoding', 'UTF8')
    kwargs = {
        'sep': hints.get('field-delimiter', ','),
        'encoding': python_encoding_from_hint.get(encoding_hint, encoding_hint),
        'header': header,
        'compression': pandas_compression_from_hint[compression_hint],
        'escapechar': hints.get('escape'),
        'iterator': True,
        'engine': 'python'
    }
    if header is None:
        # Pandas only accepts the prefix argument (which makes for
        # tidier column names when otherwise not provided) when the
        # header is explicitly marked as missing, not when it's
        # available or even when we ask Pandas to infer it.  Bummer,
        # as this means that when Pandas infers that there's no
        # header, the column names will end up different than folks
        # explicitly tell records mover that there is no header.
        #
        # https://github.com/pandas-dev/pandas/issues/27394
        # https://github.com/pandas-dev/pandas/pull/31383
        kwargs['prefix'] = 'untitled_'
    if 'quoting' in hints:
        quoting = hints['quoting']
        kwargs['quoting'] = pandas_quoting_from_hint[quoting]
    # The streaming code from pandas demands a text stream if we're
    # dealing with an uncompressed CSV file, and a binary stream
    # if we're dealing with compressed file.
    if isinstance(filepath_or_buffer, str):
        # Send a local filename...
        out = None
        try:
            out = read_csv(filepath_or_buffer, **kwargs)
            yield out
        finally:
            if out is not None:
                out.close()
    elif compression_hint is None:
        hint_encoding = hints.get('encoding')
        encoding = python_encoding_from_hint.get(hint_encoding, hint_encoding)
        text_fileobj = io.TextIOWrapper(filepath_or_buffer, encoding=encoding)
        out = None
        try:
            out = read_csv(text_fileobj, **kwargs)
            yield out
        finally:
            if out is not None:
                out.close()
            text_fileobj.detach()
    else:
        # Pass as a binary file
        out = None
        try:
            out = read_csv(filepath_or_buffer, **kwargs)
            yield out
        finally:
            if out is not None:
                out.close()
